from ase import Atoms
from gpaw import GPAW, FermiDirac


def test_symmetry_usesymm():
    a = 2.0
    L = 5.0

    a1 = Atoms('H', pbc=(1, 0, 0), cell=(a, L, L))
    a1.center()
    atoms = a1.repeat((4, 1, 1))

    def energy(symm={}):
        atoms.calc = GPAW(h=0.3,
                          occupations=FermiDirac(width=0.1),
                          symmetry=symm,
                          convergence=dict(energy=1e-6),
                          kpts=(3, 1, 1),
                          mode='lcao')
        return atoms.get_potential_energy()

    e1 = energy()
    e2 = energy('off')

    print(e1)
    print(e2)

    assert abs(e2 - e1) < 1e-6



"""
A class for finding optimal
orbitals of the KS-DFT or PZ-SIC
functionals using exponential transformation
direct minimization in lcao mode

arXiv:2101.12597 [physics.comp-ph]
Comput. Phys. Commun. 267, 108047 (2021).
https://doi.org/10.1016/j.cpc.2021.108047
"""

import numpy as np
from gpaw.lcao.eigensolver import DirectLCAO
from gpaw.utilities.tools import tri2full


class DirectMinLCAO(DirectLCAO):

    def __init__(self, wfs, ham, nkpts, diagonalizer=None,
                 orthonormalization='gramschmidt',
                 need_init_orbs=True):

        super(DirectMinLCAO, self).__init__(diagonalizer)
        super(DirectMinLCAO, self).initialize(wfs.gd, wfs.dtype,
                                              wfs.setups.nao, wfs.ksl)
        self.orthonormalization = orthonormalization
        self.need_init_orbs = need_init_orbs
        self.nkpts = nkpts
        self.reference_orbitals = {}
        self.initialize_orbitals(wfs, ham)

    def __repr__(self):
        pass

    def set_reference_orbitals(self, wfs, n_dim):
        for kpt in wfs.kpt_u:
            u = self.kpointval(kpt)
            self.reference_orbitals[u] = np.copy(kpt.C_nM[:n_dim[u]])

    def appy_transformation_kpt(self, wfs, u_mat, kpt, c_ref=None,
                                broadcast=True,
                                update_proj=True):
        """
        If c_ref are not provided then
        kpt.C_nM <- u_mat kpt.C_nM
        otherwise kpt.C_nM <- u_mat c_ref
        """

        dimens1 = u_mat.shape[1]
        dimens2 = u_mat.shape[0]

        if c_ref is None:
            kpt.C_nM[:dimens2] = u_mat @ kpt.C_nM[:dimens1]
        else:
            kpt.C_nM[:dimens2] = u_mat @ c_ref[:dimens1]

        if broadcast:
            with wfs.timer('Broadcast coefficients'):
                wfs.gd.comm.broadcast(kpt.C_nM, 0)
        if update_proj:
            with wfs.timer('Calculate projections'):
                wfs.atomic_correction.calculate_projections(wfs, kpt)

    def initialize_orbitals(self, wfs, ham):

        """
        If it is the first use of the scf then initialize
        coefficient matrix using eigensolver
        and then localise orbitals
        """

        # if it is the first use of the scf then initialize
        # coefficient matrix using eigensolver
        orthname = self.orthonormalization
        need_canon_coef = \
            (not wfs.coefficients_read_from_file and self.need_init_orbs)
        if need_canon_coef or orthname == 'diag':
            super(DirectMinLCAO, self).iterate(ham, wfs)
        else:
            wfs.orthonormalize(type=orthname)
        wfs.coefficients_read_from_file = False
        self.need_init_orbs = False

    def calc_grad(self, wfs, ham, kpt, func, evecs, evals, matrix_exp,
                  representation, ind_up, constraints):

        """
        Gradient w.r.t. skew-Hermitian matrices
        """

        h_mm = self.calculate_hamiltonian_matrix(ham, wfs, kpt)
        # make matrix hermitian
        tri2full(h_mm)
        # calc gradient and eigenstate error
        g_mat, error = func.get_gradients(
            h_mm, kpt.C_nM, kpt.f_n, evecs, evals,
            kpt, wfs, wfs.timer, matrix_exp,
            representation, ind_up, constraints)

        return g_mat, error

    def update_to_canonical_orbitals(self, wfs, ham, kpt,
                                     update_ref_orbs_canonical, restart):
        """
        Choose canonical orbitals
        """

        h_mm = self.calculate_hamiltonian_matrix(ham, wfs, kpt)
        tri2full(h_mm)
        if update_ref_orbs_canonical or restart:
            # Diagonalize entire Hamiltonian matrix
            with wfs.timer('Diagonalize and rotate'):
                kpt.C_nM, kpt.eps_n = rotate_subspace(h_mm, kpt.C_nM)
        else:
            # Diagonalize equally occupied subspaces separately
            f_unique = np.unique(kpt.f_n)
            for f in f_unique:
                with wfs.timer('Diagonalize and rotate'):
                    kpt.C_nM[kpt.f_n == f, :], kpt.eps_n[kpt.f_n == f] = \
                        rotate_subspace(h_mm, kpt.C_nM[kpt.f_n == f, :])

        with wfs.timer('Calculate projections'):
            self.update_projections(wfs, kpt)

    def sort_orbitals(self, wfs, kpt, ind):
        """
        sort orbitals according to indices stored in ind
        """
        kpt.C_nM[np.arange(len(ind)), :] = kpt.C_nM[ind, :]
        self.update_projections(wfs, kpt)

    def update_projections(self, wfs, kpt):
        """
        calculate projections kpt.P_ani
        """

        wfs.atomic_correction.calculate_projections(wfs, kpt)

    def orbital_energies(self, wfs, ham, kpt):
        """
        diagonal elements of hamiltonian matrix in orbital representation
        """

        h_mm = self.calculate_hamiltonian_matrix(ham, wfs, kpt)
        tri2full(h_mm)
        # you probably need only diagonal terms?
        # wouldn't "for" be faster?
        h_mm = kpt.C_nM.conj() @ h_mm.conj() @ kpt.C_nM.T

        return h_mm.diagonal().real.copy()

    def kpointval(self, kpt):
        return self.nkpts * kpt.s + kpt.q


def rotate_subspace(h_mm, c_nm):
    """
    choose canonical orbitals
    """
    l_nn = c_nm @ h_mm @ c_nm.conj().T
    # check if diagonal then don't rotate? it could save a bit of time
    eps, w = np.linalg.eigh(l_nn)
    return w.T.conj() @ c_nm, eps



"""Calculate diamond with various parallelizations with GLLBSC."""
import pytest
from gpaw.mpi import world
from ase.build import bulk
from gpaw import GPAW, Mixer


@pytest.mark.skipif(world.size < 4,
                    reason='world.size < 4')
def test_parallel_diamond_gllb(in_tmp_dir):
    xc = 'GLLBSC'
    KS_gap_ref = 4.180237125868162
    QP_gap_ref = 5.469387490357182
    # M. Kuisma et. al, https://doi.org/10.1103/PhysRevB.82.115106
    #     C: KS gap 4.14 eV, QP gap 5.41eV, expt. 5.48 eV
    KSb = []
    dxcb = []

    eigensolver = 'rmm-diis'

    for band in [1, 2, 4]:
        # Calculate ground state
        atoms = bulk('C', 'diamond', a=3.567)
        calc = GPAW(h=0.15,
                    kpts=(4, 4, 4),
                    xc=xc,
                    nbands=8,
                    mixer=Mixer(0.5, 5, 50.0),
                    eigensolver=eigensolver,
                    parallel={'band': band})
        atoms.calc = calc
        atoms.get_potential_energy()

        # Calculate accurate KS-band gap from band structure
        bs_calc = calc.fixed_density(kpts={'path': 'GX', 'npoints': 12},
                                     symmetry='off',
                                     nbands=8,
                                     convergence={'bands': 8},
                                     eigensolver=eigensolver)
        # Get the accurate KS-band gap
        homo, lumo = bs_calc.get_homo_lumo()

        # Calculate the discontinuity potential with accurate band gap
        response = calc.hamiltonian.xc.response
        dxc_pot = response.calculate_discontinuity_potential(homo, lumo)

        # Calculate the discontinuity using the band structure calculator
        bs_response = bs_calc.hamiltonian.xc.response
        KS_gap, dxc = bs_response.calculate_discontinuity(dxc_pot)
        assert KS_gap == pytest.approx(lumo - homo, abs=1e-10)
        assert KS_gap == pytest.approx(KS_gap_ref, abs=1e-4)

        QP_gap = KS_gap + dxc
        assert QP_gap == pytest.approx(QP_gap_ref, abs=1e-4)
        KSb.append(KS_gap)
        dxcb.append(dxc)

    assert abs(KSb[0] - KSb[1]) < 1e-6
    assert abs(KSb[0] - KSb[2]) < 1e-6
    assert abs(dxcb[0] - dxcb[1]) < 1e-6
    assert abs(dxcb[0] - dxcb[2]) < 1e-6



import pytest
import numpy as np
import pickle
from ase.build import molecule
from gpaw.lcao.tools import makeU, makeV
from gpaw import GPAW, FermiDirac, restart
from gpaw.lcao.pwf2 import LCAOwrap
from gpaw.mpi import world, rank, serial_comm
from gpaw.test import equal

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


@pytest.mark.later
def test_lcao_pair_and_coulomb(in_tmp_dir):
    atoms = molecule('H2')
    atoms.set_cell([6.4, 6.4, 6.4])
    atoms.center()
    calc = GPAW(mode='lcao', occupations=FermiDirac(0.1),
                poissonsolver={'name': 'fd'})
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('lcao_pair.gpw')

    if rank == 0:
        atoms, calc = restart('lcao_pair.gpw',
                              txt=None, communicator=serial_comm)
        lcao = LCAOwrap(calc)
        H = lcao.get_hamiltonian()
        S = lcao.get_overlap()
        with open('lcao_pair_hs.pckl', 'wb') as fd:
            pickle.dump((H, S), fd, 2)
        # indices = get_bfi2(symbols, basis, scat)
        indices = range(2)
        lcao.get_xc(indices=indices).dump('lcao_pair_xc.pckl')
        lcao.get_Fcore(indices=indices).dump('lcao_pair_Fcore.pckl')
        w_wG = lcao.get_orbitals(indices=indices)
        P_awi = lcao.get_projections(indices=indices)
        with open('lcao_pair_w_wG__P_awi.pckl', 'wb') as fd:
            pickle.dump((w_wG, P_awi), fd, 2)

    world.barrier()
    makeU('lcao_pair.gpw',
          'lcao_pair_w_wG__P_awi.pckl',
          'lcao_pair_eps_q__U_pq.pckl',
          1.0e-5,
          dppname='lcao_pair_D_pp.pckl')

    world.barrier()
    makeV('lcao_pair.gpw',
          'lcao_pair_w_wG__P_awi.pckl',
          'lcao_pair_eps_q__U_pq.pckl',
          'lcao_pair_V_qq.pckl',
          'lcao_pair_V_qq.log',
          False)

    world.barrier()
    V_qq = np.load('lcao_pair_V_qq.pckl', allow_pickle=True)
    eps_q, U_pq = np.load('lcao_pair_eps_q__U_pq.pckl', allow_pickle=True)
    assert U_pq.flags.contiguous
    Usq_pq = U_pq * np.sqrt(eps_q)
    V_pp = np.dot(np.dot(Usq_pq, V_qq), Usq_pq.T.conj())
    V_pp_ref = np.array(
        [[15.34450177, 11.12669608, 11.12669608, 12.82934563],
         [11.12669608, 8.82280293, 8.82280293, 11.12669608],
         [11.12669608, 8.82280293, 8.82280293, 11.12669608],
         [12.82934563, 11.12669608, 11.12669608, 15.34450178]])
    equal(abs(V_pp_ref - V_pp).max(), 0.0, 1.0e-5)



import numpy as np
import pytest

from gpaw.response.g0w0 import G0W0


reference_kn = [[0.69806561, 2.58472004, 2.58472066,
                 2.58469313, 3.60859825, 3.60859883],
                [0.87642735, 1.02930988, 4.52049808,
                 4.85337269, 4.85355968, 9.60323838],
                [0.96375991, 2.57490687, 2.57494555,
                 4.59771405, 4.59774543, 8.67625318]]


@pytest.mark.later  # Need to implement Hubbard U first
@pytest.mark.response
def test_hubbard_GW(in_tmp_dir, gpw_files, needs_ase_master):
    # This tests checks the actual numerical accuracy which is asserted below
    gw = G0W0(gpw_files['ag_plusU_pw_wfs'], 'gw',
              integrate_gamma=0,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.1, 'omegamax': None},
              nbands=19,  # Carefully selected to avoid slicing degenerate band
              ecut=52.8,  # This too
              eta=0.2)
    results = gw.calculate()

    qp_kn = results['qp'][0]

    assert np.allclose(qp_kn, reference_kn, atol=1e-4, rtol=1e-4)



import numpy as np
from gpaw.solvation.gridmem import NeedsGD


class Dielectric(NeedsGD):
    """Class representing a spatially varying permittivity.

    Attributes:
    eps_gradeps     -- List [eps_g, dxeps_g, dyeps_g, dzeps_g]
                       with
                       eps_g:   permittivity on the fine grid
                       dieps_g: gradient of eps_g.
    del_eps_del_g_g -- Partial derivative with respect to
                       the cavity. (A point-wise dependence
                       on the cavity is assumed).
    """

    def __init__(self, epsinf):
        """Constructor for the Dielectric class.

        Arguments:
        epsinf -- Static dielectric constant
                  at infinite distance from the solute.
        """
        NeedsGD.__init__(self)
        self._epsinf = float(epsinf)
        self.eps_gradeps = None  # eps_g, dxeps_g, dyeps_g, dzeps_g
        self.del_eps_del_g_g = None
        self.cavity = None

    def estimate_memory(self, mem):
        nbytes = self.gd.bytecount()
        mem.subnode('Permittivity', nbytes)
        mem.subnode('Permittivity Gradient', 3 * nbytes)
        mem.subnode('Permittivity Derivative', nbytes)

    def allocate(self):
        NeedsGD.allocate(self)
        self.eps_gradeps = []
        eps_g = self.gd.empty()
        eps_g.fill(1.0)
        self.eps_gradeps.append(eps_g)
        self.eps_gradeps.extend([gd.zeros() for gd in (self.gd, ) * 3])
        self.del_eps_del_g_g = self.gd.empty()

    @property
    def epsinf(self):
        return self._epsinf

    @epsinf.setter
    def epsinf(self, epsinf):
        if epsinf != self._epsinf:
            self._epsinf = float(epsinf)
            self.del_eps_del_g_g = self._epsinf - 1.
            if self.cavity is not None:
                self.update(self.cavity)

    def update(self, cavity):
        """Calculate eps_gradeps and del_eps_del_g_g from the cavity."""
        self.cavity = cavity
        self.update_eps_only()
        for i in (0, 1, 2):
            np.multiply(
                self.del_eps_del_g_g,
                cavity.grad_g_vg[i],
                self.eps_gradeps[1 + i]
            )

    def update_eps_only(self):
        raise NotImplementedError

    def __str__(self):
        s = 'Dielectric: %s\n' % (self.__class__, )
        s += '  epsilon_inf: %s\n' % (self._epsinf, )
        return s

    def write(self, writer):
        writer.write(name=self.__class__.__name__, epsinf=self._epsinf)


class LinearDielectric(Dielectric):
    """Dielectric depending (affine) linearly on the cavity.

    See also
    A. Held and M. Walter, J. Chem. Phys. 141, 174108 (2014).
    """
    def allocate(self):
        Dielectric.allocate(self)
        self.del_eps_del_g_g = self._epsinf - 1.  # frees array

    def update_eps_only(self):
        np.multiply(self.cavity.g_g, self._epsinf - 1., self.eps_gradeps[0])
        self.eps_gradeps[0] += 1.

    def write(self, writer):
        writer.write(name='LinearDielectric',
                     epsinf=self.epsinf)


class CMDielectric(Dielectric):
    """Clausius-Mossotti like dielectric.

    Untested, use at own risk!
    """

    def update_eps_only(self):
        ei = self._epsinf
        t = 1. - self.cavity.g_g
        self.eps_gradeps[0][:] = (3. * (ei + 2.)) / ((ei - 1.) * t + 3.) - 2.
        self.del_eps_del_g_g[:] = (
            (3. * (ei - 1.) * (ei + 2.)) / ((ei - 1.) * t + 3.) ** 2
        )



import pytest
from gpaw.xc import XC


@pytest.mark.libxc
def test_pbe0():
    xc = XC('PBE0:backend=pw')
    assert xc.exx_fraction == 0.25
    xc = XC('HYB_GGA_XC_PBEH:backend=pw:omega=0:fraction=0.25')
    assert xc.exx_fraction == 0.25



import numpy as np
from scipy.special import erf

from gpaw.atom.atompaw import AtomPAW
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.basis_data import Basis, BasisFunction
from gpaw.setup import BaseSetup, LocalCorrectionVar
from gpaw.spline import Spline
from gpaw.utilities import divrl, hartree as hartree_solve


null_spline = Spline(0, 1.0, [0., 0., 0.])


# XXX Not used at the moment; see comment below about rgd splines.
def projectors_to_splines(rgd, l_j, pt_jg, filter=None):
    # This function exists because both HGH and SG15 needs to do
    # exactly the same thing.
    #
    # XXX equal-range projectors still required for some reason
    maxlen = max([len(pt_g) for pt_g in pt_jg])
    pt_j = []
    for l, pt1_g in zip(l_j, pt_jg):
        pt2_g = np.zeros(maxlen)
        pt2_g[:len(pt1_g)] = pt1_g
        if filter is not None:
            filter(rgd, rgd.r_g[maxlen], pt2_g, l=l)
        pt2_g = divrl(pt2_g, l, rgd.r_g[:maxlen])
        spline = rgd.spline(pt2_g, rgd.r_g[maxlen - 1], l=l)
        pt_j.append(spline)
    return pt_j


# XXX not used at the moment
def local_potential_to_spline(rgd, vbar_g, filter=None):
    vbar_g = vbar_g.copy()
    rcut = rgd.r_g[len(vbar_g) - 1]
    if filter is not None:
        filter(rgd, rcut, vbar_g, l=0)
    # vbar = Spline(0, rcut, vbar_g)
    vbar = rgd.spline(vbar_g, rgd.r_g[len(vbar_g) - 1], l=0)
    return vbar


def get_radial_hartree_energy(r_g, rho_g):
    """Get energy of l=0 compensation charge on equidistant radial grid."""

    # At least in some cases the zeroth point is moved to 1e-8 or so to
    # prevent division by zero and the like, so:
    dr = r_g[2] - r_g[1]
    rho_r_dr_g = dr * r_g * rho_g
    vh_r_g = np.zeros(len(r_g))  # "r * vhartree"
    hartree_solve(0, rho_r_dr_g, r_g, vh_r_g)
    return 2.0 * np.pi * (rho_r_dr_g * vh_r_g).sum()


def screen_potential(r, v, charge, rcut=None, a=None):
    """Split long-range potential into short-ranged contributions.

    The potential v is a long-ranted potential with the asymptotic form Z/r
    corresponding to the given charge.

    Return a potential vscreened and charge distribution rhocomp such that

      v(r) = vscreened(r) + vHartree[rhocomp](r).

    The returned quantities are truncated to a reasonable cutoff radius.
    """
    vr = v * r + charge

    if rcut is None:
        err = 0.0
        i = len(vr)
        while err < 1e-4:
            # Things can be a bit sensitive to the threshold.  The O.pz-mt
            # setup gets 20-30 Bohr long compensation charges if it's 1e-6.
            i -= 1
            err = abs(vr[i])
        i += 1

        icut = np.searchsorted(r, r[i] * 1.1)
    else:
        icut = np.searchsorted(r, rcut)
    rcut = r[icut]
    rshort = r[:icut].copy()
    if rshort[0] < 1e-16:
        rshort[0] = 1e-10

    if a is None:
        a = rcut / 5.0  # XXX why is this so important?
    vcomp = np.zeros_like(rshort)
    vcomp = charge * erf(rshort / (np.sqrt(2.0) * a)) / rshort
    # XXX divide by r
    rhocomp = charge * (np.sqrt(2.0 * np.pi) * a)**(-3) * \
        np.exp(-0.5 * (rshort / a)**2)
    vscreened = v[:icut] + vcomp
    return vscreened, rhocomp


def figure_out_valence_states(ppdata):
    from gpaw.atom.configurations import configurations
    from ase.data import chemical_symbols
    # ppdata.symbol may not be a chemical symbol so use Z
    chemical_symbol = chemical_symbols[ppdata.Z]
    Z, config = configurations[chemical_symbol]
    assert Z == ppdata.Z

    # Okay, we need to figure out occupations f_ln when we don't know
    # any info about existing states on the pseudopotential.
    #
    # The plan is to loop over all states and count until only the correct
    # number of valence electrons "remain".
    nelectrons = 0
    ncore = ppdata.Z - ppdata.Nv

    energies = [c[3] for c in config]
    args = np.argsort(energies)
    config = list(np.array(config, dtype=object)[args])

    nelectrons = 0
    ncore = ppdata.Z - ppdata.Nv
    assert ppdata.Nv > 0
    iterconfig = iter(config)
    if ncore > 0:
        for n, l, occ, eps in iterconfig:
            nelectrons += occ
            if nelectrons == ncore:
                break
            elif nelectrons >= ncore:
                raise ValueError('Cannot figure out what states should exist '
                                 'on this pseudopotential.')

    f_ln = {}
    l_j = []
    f_j = []
    n_j = []
    for n, l, occ, eps in iterconfig:
        f_ln.setdefault(l, []).append(occ)
        l_j.append(l)
        f_j.append(occ)
        n_j.append(n)
    lmax = max(f_ln.keys())
    f_ln = [f_ln.get(l, []) for l in range(lmax + 1)]
    return n_j, l_j, f_j, f_ln


def generate_basis_functions(ppdata):
    class SimpleBasis(Basis):
        def __init__(self, symbol, l_j):
            rgd = EquidistantRadialGridDescriptor(0.02, 160)
            Basis.__init__(self, symbol, 'simple', readxml=False, rgd=rgd)
            self.generatordata = 'simple'
            bf_j = self.bf_j
            rcgauss = rgd.r_g[-1] / 3.0
            gauss_g = np.exp(-(rgd.r_g / rcgauss)**2.0)
            for l in l_j:
                phit_g = rgd.r_g**l * gauss_g
                norm = (rgd.integrate(phit_g**2) / (4 * np.pi))**0.5
                phit_g /= norm
                bf = BasisFunction(None, l, rgd.r_g[-1], phit_g, 'gaussian')
                bf_j.append(bf)
    # l_orb_J = [state.l for state in self.data['states']]
    b1 = SimpleBasis(ppdata.symbol, ppdata.l_orb_J)
    apaw = AtomPAW(ppdata.symbol, [ppdata.f_ln], h=0.05, rcut=9.0,
                   basis={ppdata.symbol: b1},
                   setups={ppdata.symbol: ppdata},
                   maxiter=60,
                   txt=None)
    basis = apaw.extract_basis_functions()
    return basis


def pseudoplot(pp, show=True):
    import matplotlib.pyplot as plt

    fig = plt.figure()
    wfsax = fig.add_subplot(221)
    ptax = fig.add_subplot(222)
    vax = fig.add_subplot(223)
    rhoax = fig.add_subplot(224)

    def spline2grid(spline):
        rcut = spline.get_cutoff()
        r = np.linspace(0.0, rcut, 2000)
        return r, spline.map(r)

    for phit in pp.phit_j:
        r, y = spline2grid(phit)
        wfsax.plot(r, y, label='wf l=%d' % phit.get_angular_momentum_number())

    for pt in pp.pt_j:
        r, y = spline2grid(pt)
        ptax.plot(r, y, label='pr l=%d' % pt.get_angular_momentum_number())

    for ghat in pp.ghat_l:
        r, y = spline2grid(ghat)
        rhoax.plot(r, y, label='cc l=%d' % ghat.get_angular_momentum_number())

    r, y = spline2grid(pp.vbar)
    vax.plot(r, y, label='vbar')

    vax.set_ylabel('potential')
    rhoax.set_ylabel('density')
    wfsax.set_ylabel('wfs')
    ptax.set_ylabel('projectors')

    for ax in [vax, rhoax, wfsax, ptax]:
        ax.legend()

    if show:
        plt.show()


class PseudoPotential(BaseSetup):
    def __init__(self, data, basis=None, filter=None):
        self.data = data

        self.R_sii = None
        self.lq = None

        self.filename = None
        self.fingerprint = None
        self.symbol = data.symbol
        self.type = data.name

        self.Z = data.Z
        self.Nv = data.Nv
        self.Nc = data.Nc

        self.f_j = data.f_j
        self.n_j = data.n_j
        self.l_j = data.l_j
        self.l_orb_J = data.l_orb_J
        self.nj = len(data.l_j)

        self.ni = sum([2 * l + 1 for l in data.l_j])
        # self.pt_j = projectors_to_splines(data.rgd, data.l_j, data.pt_jg,
        #                                   filter=filter)
        self.pt_j = data.get_projectors()

        if len(self.pt_j) == 0:
            assert False  # not sure yet about the consequences of
            # cleaning this up in the other classes
            self.l_j = [0]
            self.pt_j = [null_spline]

        if basis is None:
            basis = data.create_basis_functions()

        self.basis_functions_J = basis.tosplines()

        # We declare (for the benefit of the wavefunctions reuse method)
        # that we have no PAW projectors as such.  This makes the
        # 'paw' wfs reuse method a no-op.
        self.pseudo_partial_waves_j = []

        self.basis = basis
        self.nao = sum([2 * phit.get_angular_momentum_number() + 1
                        for phit in self.basis_functions_J])

        self.Nct = 0.0
        self.nct = null_spline

        self.lmax = 0

        self.xc_correction = None

        r, l_comp, g_comp = data.get_compensation_charge_functions()
        assert l_comp == [0]  # Presumably only spherical charges
        self.ghat_l = [Spline(l, r[-1], g) for l, g in zip(l_comp, g_comp)]
        self.rcgauss = data.rcgauss

        # accuracy is rather sensitive to this
        # self.vbar = local_potential_to_spline(data.rgd, data.vbar_g,
        #                                      filter=filter)
        self.vbar = data.get_local_potential()
        # XXX HGH and UPF use different radial grids, and this for
        # some reason makes it difficult to use the exact same code to
        # construct vbar and projectors.  This should be fixed since
        # either type of rgd should be able to always produce a valid
        # and equivalent spline transparently.

        _np = self.ni * (self.ni + 1) // 2
        self.Delta0 = data.Delta0
        self.Delta_pL = np.zeros((_np, 1))

        self.E = 0.0
        self.Kc = 0.0
        self.M = -data.Eh_compcharge
        self.M_p = np.zeros(_np)
        self.M_pp = np.zeros((_np, _np))
        self.M_wpp = {}

        self.K_p = data.expand_hamiltonian_matrix()
        self.MB = 0.0
        self.MB_p = np.zeros(_np)
        self.dO_ii = np.zeros((self.ni, self.ni))

        # We don't really care about these variables
        self.rcutfilter = None
        self.rcore = None

        self.N0_p = np.zeros(_np)  # not really implemented
        self.nabla_iiv = None
        self.rxnabla_iiv = None
        self.phicorehole_g = None
        self.rgd = data.rgd
        self.rcut_j = data.rcut_j
        self.tauct = None
        self.Delta_iiL = np.zeros((self.ni, self.ni, 1))
        self.B_ii = None
        self.dC_ii = None
        self.X_p = None
        self.X_wp = {}
        self.X_pg = None
        self.ExxC = None
        self.ExxC_w = {}
        self.X_gamma = None
        self.dEH0 = 0.0
        self.dEH_p = np.zeros(_np)
        self.extra_xc_data = {}

        self.wg_lg = None
        self.g_lg = None
        self.local_corr = LocalCorrectionVar(None)
        self._Mg_pp = None
        self._gamma = None



"""Code for reading old gpw files."""
import warnings

import numpy as np
import ase.io.ulm as ulm
from ase import Atoms
from ase.dft.kpoints import monkhorst_pack
from ase.io.trajectory import write_atoms
from ase.units import AUT, Bohr, Ha
from gpaw.utilities import devnull

from gpaw import PW
from gpaw.occupations import FermiDirac
from gpaw.io.tar import Reader


def wrap_old_gpw_reader(filename):
    warnings.warn('You are reading an old-style gpw-file.  Please check '
                  'the results carefully!')

    r = Reader(filename)

    data = {'version': -1,
            'gpaw_version': '1.0',
            'ha': Ha,
            'bohr': Bohr,
            'scf.': {'converged': True},
            'atoms.': {},
            'wave_functions.': {}}

    class DictBackend:
        def write(self, **kwargs):
            data['atoms.'].update(kwargs)

    write_atoms(DictBackend(), read_atoms(r))

    e_total_extrapolated = r.get('PotentialEnergy').item() * Ha
    magmom_a = r.get('MagneticMoments')
    data['results.'] = {
        'energy': e_total_extrapolated,
        'magmoms': magmom_a,
        'magmom': magmom_a.sum()}

    if r.has_array('CartesianForces'):
        data['results.']['forces'] = r.get('CartesianForces') * Ha / Bohr

    p = data['parameters.'] = {}

    p['xc'] = r['XCFunctional']
    p['nbands'] = r.dimension('nbands')
    p['spinpol'] = (r.dimension('nspins') == 2)

    bzk_kc = r.get('BZKPoints', broadcast=True)
    if r.has_array('NBZKPoints'):
        p['kpts'] = r.get('NBZKPoints', broadcast=True)
        if r.has_array('MonkhorstPackOffset'):
            offset_c = r.get('MonkhorstPackOffset', broadcast=True)
            if offset_c.any():
                p['kpts'] = monkhorst_pack(p['kpts']) + offset_c
    else:
        p['kpts'] = bzk_kc

    if r['version'] < 4:
        usesymm = r['UseSymmetry']
        if usesymm is None:
            p['symmetry'] = {'time_reversal': False, 'point_group': False}
        elif usesymm:
            p['symmetry'] = {'time_reversal': True, 'point_group': True}
        else:
            p['symmetry'] = {'time_reversal': True, 'point_group': False}
    else:
        p['symmetry'] = {'point_group': r['SymmetryOnSwitch'],
                         'symmorphic': r['SymmetrySymmorphicSwitch'],
                         'time_reversal': r['SymmetryTimeReversalSwitch'],
                         'tolerance': r['SymmetryToleranceCriterion']}

    p['basis'] = r['BasisSet']

    try:
        h = r['GridSpacing']
    except KeyError:  # CMR can't handle None!
        h = None
    if h is not None:
        p['h'] = Bohr * h
    if r.has_array('GridPoints'):
        p['gpts'] = r.get('GridPoints')

    p['lmax'] = r['MaximumAngularMomentum']
    p['setups'] = r['SetupTypes']
    p['fixdensity'] = r['FixDensity']
    nbtc = r['NumberOfBandsToConverge']
    if not isinstance(nbtc, (int, str)):
        # The string 'all' was eval'ed to the all() function!
        nbtc = 'all'
    p['convergence'] = {'density': r['DensityConvergenceCriterion'],
                        'energy': r['EnergyConvergenceCriterion'] * Ha,
                        'eigenstates': r['EigenstatesConvergenceCriterion'],
                        'bands': nbtc}
    mixer = r['MixClass']
    weight = r['MixWeight']

    for key in ['basis', 'setups']:
        dct = p[key]
        if isinstance(dct, dict) and None in dct:
            dct['default'] = dct.pop(None)

    if mixer == 'Mixer':
        from gpaw.mixer import Mixer
    elif mixer == 'MixerSum':
        from gpaw.mixer import MixerSum as Mixer
    elif mixer == 'MixerSum2':
        from gpaw.mixer import MixerSum2 as Mixer
    elif mixer == 'MixerDif':
        from gpaw.mixer import MixerDif as Mixer
    elif mixer == 'DummyMixer':
        from gpaw.mixer import DummyMixer as Mixer
    else:
        Mixer = None

    if Mixer is None:
        p['mixer'] = None
    else:
        p['mixer'] = Mixer(r['MixBeta'], r['MixOld'], weight)

    p['stencils'] = (r['KohnShamStencil'],
                     r['InterpolationStencil'])

    vt_sG = r.get('PseudoPotential') * Ha
    ps = r['PoissonStencil']
    if isinstance(ps, int) or ps == 'M':
        poisson = {'name': 'fd'}
        poisson['nn'] = ps
        if data['atoms.']['pbc'] == [1, 1, 0]:
            v1, v2 = vt_sG[0, :, :, [0, -1]].mean(axis=(1, 2))
            if abs(v1 - v2) > 0.01:
                warnings.warn('I am guessing that this calculation was done '
                              'with a dipole-layer correction?')
                poisson['dipolelayer'] = 'xy'
        p['poissonsolver'] = poisson

    p['charge'] = r['Charge']
    fixmom = r['FixMagneticMoment']

    p['occupations'] = FermiDirac(r['FermiWidth'] * Ha,
                                  fixmagmom=fixmom)

    p['mode'] = r['Mode']

    if p['mode'] == 'pw':
        p['mode'] = PW(ecut=r['PlaneWaveCutoff'] * Ha)

    if len(bzk_kc) == 1 and not bzk_kc[0].any():
        # Gamma point only:
        if r['DataType'] == 'Complex':
            p['dtype'] = complex

    data['occupations.'] = {
        'fermilevel': r['FermiLevel'] * Ha,
        'split': r.parameters.get('FermiSplit', 0) * Ha,
        'homo': np.nan,
        'lumo': np.nan}

    data['density.'] = {
        'density': r.get('PseudoElectronDensity') * Bohr**-3,
        'atomic_density_matrices': r.get('AtomicDensityMatrices')}

    data['hamiltonian.'] = {
        'e_coulomb': r['Epot'] * Ha,
        'e_entropy': -r['S'] * Ha,
        'e_external': r['Eext'] * Ha,
        'e_kinetic': r['Ekin'] * Ha,
        'e_total_extrapolated': e_total_extrapolated,
        'e_xc': r['Exc'] * Ha,
        'e_zero': r['Ebar'] * Ha,
        'potential': vt_sG,
        'atomic_hamiltonian_matrices': r.get('NonLocalPartOfHamiltonian') * Ha}
    data['hamiltonian.']['e_total_free'] = (
        sum(data['hamiltonian.'][e] for e in ['e_coulomb', 'e_entropy',
                                              'e_external', 'e_kinetic',
                                              'e_xc', 'e_zero']))

    if r.has_array('GLLBPseudoResponsePotential'):
        data['hamiltonian.']['xc.'] = {
            'gllb_pseudo_response_potential':
            r.get('GLLBPseudoResponsePotential') * Ha,
            'gllb_dxc_pseudo_response_potential':
            r.get('GLLBDxcPseudoResponsePotential') * Ha / Bohr,
            'gllb_atomic_density_matrices':
            r.get('GLLBAtomicDensityMatrices'),
            'gllb_atomic_response_matrices':
            r.get('GLLBAtomicResponseMatrices'),
            'gllb_dxc_atomic_density_matrices':
            r.get('GLLBDxcAtomicDensityMatrices'),
            'gllb_dxc_atomic_response_matrices':
            r.get('GLLBDxcAtomicResponseMatrices')}

    special = [('eigenvalues', 'Eigenvalues'),
               ('occupations', 'OccupationNumbers'),
               ('projections', 'Projections')]

    if r['Mode'] == 'pw':
        special.append(('coefficients', 'PseudoWaveFunctions'))
        try:
            data['wave_functions.']['indices'] = r.get('PlaneWaveIndices')
        except KeyError:
            pass
    elif r['Mode'] == 'fd':
        special.append(('values', 'PseudoWaveFunctions'))
    else:
        special.append(('coefficients', 'WaveFunctionCoefficients'))

    for name, old in special:
        try:
            fd, shape, size, dtype = r.get_file_object(old, ())
        except KeyError:
            continue
        offset = fd
        data['wave_functions.'][name + '.'] = {
            'ndarray': (shape, dtype.name, offset)}

    new = ulm.Reader(devnull, data=data,
                     little_endian=r.byteswap ^ np.little_endian)

    for ref in new._data['wave_functions']._data.values():
        try:
            ref.fd = ref.offset
        except AttributeError:
            continue
        ref.offset = 0

    return new


def read_atoms(reader):
    positions = reader.get('CartesianPositions', broadcast=True) * Bohr
    numbers = reader.get('AtomicNumbers', broadcast=True)
    cell = reader.get('UnitCell', broadcast=True) * Bohr
    pbc = reader.get('BoundaryConditions', broadcast=True)
    tags = reader.get('Tags', broadcast=True)
    magmoms = reader.get('MagneticMoments', broadcast=True)

    # Create instance of Atoms object, and set_tags and magnetic moments
    atoms = Atoms(positions=positions,
                  numbers=numbers,
                  cell=cell,
                  pbc=pbc)

    if tags.any():
        atoms.set_tags(tags)
    if magmoms.any():
        atoms.set_initial_magnetic_moments(magmoms)

    if reader.has_array('CartesianVelocities'):
        velocities = reader.get('CartesianVelocities',
                                broadcast=True) * Bohr / AUT
        atoms.set_velocities(velocities)

    return atoms



import pytest
from gpaw.lrtddft import d2Excdnsdnt, d2Excdn2
import numpy as np


@pytest.mark.skip(reason='TODO')
def test_d2Excdn2():
    dup = np.array([.1, .3])
    ddn = np.array([.2, .005])

    def equiv(xl, yl):
        for x, y in zip(xl, yl):
            #        print x,y
            assert abs(x - y) < 1e-8

    out = 0

    # unpolarised ......................................

    expected = [-0.99599492, -12.51604312]
    res = d2Excdn2(ddn)
    if out:
        print(res)
    equiv(xl=res, yl=expected)

    # polarised ........................................

    isp = 1
    ksp = 1
    expected = [-1.6505808, -0.93437538]
    res = d2Excdnsdnt(dup, ddn)
    if out:
        print(res)
    equiv(res[isp][ksp], expected)

    isp = 1
    ksp = 0
    expected = [-0.18324442, -0.11948321]
    res = d2Excdnsdnt(dup, ddn)
    if out:
        print(res)
    equiv(res[isp][ksp], expected)

    isp = 0
    ksp = 0
    expected = [-1.14670206, -11.02164441]
    res = d2Excdnsdnt(dup, ddn)
    if out:
        print(res)
    equiv(res[isp][ksp], expected)



import sys
from math import pi, exp, sqrt, log

import numpy as np
from scipy.optimize import fsolve
from scipy.linalg import eigh
from scipy.special import erf
from ase.units import Hartree
from ase.data import atomic_numbers, chemical_symbols

from gpaw import __version__ as version
from gpaw.basis_data import Basis, BasisFunction, BasisPlotter
from gpaw.gaunt import gaunt
from gpaw.utilities import pack2
from gpaw.atom.aeatom import (AllElectronAtom, Channel, parse_ld_str, colors,
                              GaussianBasis)
from gpaw.xc.ri.ribasis import generate_ri_basis
from gpaw.xc.ri.spherical_hse_kernel import RadialHSE


class DatasetGenerationError(Exception):
    pass


parameters = {
    # 1-2:
    'H1': ('1s,s,p', 0.9),
    'He2': ('1s,s,p', 1.5),
    # 3-10:
    'Li1': ('2s,s,2p', 2.1),
    'Li3': ('1s,2s,2p,p,d', 1.5),
    'Be2': ('2s,s,2p', 1.5),
    'Be4': ('1s,2s,2p,p,d', 1.4),
    'B3': ('2s,s,2p,p,d', 1.2),
    'C4': ('2s,s,2p,p,d', 1.2),
    'N5': ('2s,s,2p,p,d', [1.2, 1.3], {'r0': 1.1}),
    'O6': ('2s,s,2p,p,d,F', 1.2),
    'F7': ('2s,s,2p,p,d', [1.2, 1.4]),
    'Ne8': ('2s,s,2p,p,d', 1.8),
    # 11-18:
    'Na1': ('3s,s,3p', 2.6),
    'Na9': ('2s,3s,2p,3p,d,F', 2.3),
    'Mg2': ('3s,s,3p,D', 2.6),
    'Mg10': ('2s,3s,2p,3p,d,F', [2.0, 1.8]),
    'Al3': ('3s,s,3p,p,d,F', 2.1),
    'Si4': ('3s,s,3p,p,d,F', 1.9),
    'P5': ('3s,s,3p,p,d,F', 1.7),
    'S6': ('3s,s,3p,p,d,F', 1.6),
    'Cl7': ('3s,s,3p,p,d,F', 1.5),
    'Ar8': ('3s,s,3p,p,d,F', 1.5),
    # 19-36:
    'K1': ('4s,s,4p,D', 3.5),
    'K9': ('3s,4s,3p,4p,d,d,F', 2.1),
    'Ca2': ('4s,s,4p', 3.1),
    'Ca10': ('3s,4s,3p,4p,3d,d,F', 2.1),
    'Sc3': ('4s,s,4p,p,3d,d', 2.7),
    'Sc11': ('3s,4s,3p,4p,3d,d,F', 2.3),
    'Ti4': ('4s,s,4p,p,3d,d', 2.7),
    'Ti12': ('3s,4s,3p,4p,3d,d,F', [2.2, 2.2, 2.3]),
    'V5': ('4s,s,4p,p,3d,d', 2.6),
    'V13': ('3s,4s,3p,4p,3d,d,F', [2.1, 2.1, 2.3]),
    'Cr6': ('4s,s,4p,p,3d,d', 2.5),
    'Cr14': ('3s,4s,3p,4p,3d,d,F', [2.1, 2.1, 2.3]),
    'Mn7': ('4s,s,4p,p,3d,d', 2.4),
    'Mn15': ('3s,4s,3p,4p,3d,d,F', [2.0, 2.0, 2.2]),
    'Fe8': ('4s,s,4p,p,3d,d', 2.2),
    'Fe16': ('3s,4s,3p,4p,3d,d,F', 2.1),
    'Co9': ('4s,s,4p,p,3d,d', 2.2),
    'Co17': ('3s,4s,3p,4p,3d,d,F', 2.1),
    'Ni10': ('4s,s,4p,p,3d,d', 2.1),
    'Ni18': ('3s,4s,3p,4p,3d,d,F', 2.0),
    'Cu11': ('4s,s,4p,p,3d,d', 2.1),
    'Cu19': ('3s,4s,3p,4p,3d,d,F', 1.9),
    'Zn12': ('4s,s,4p,p,3d', 2.1),
    'Zn20': ('3s,4s,3p,4p,3d,d,F', 1.9),
    'Ga3': ('4s,s,4p,p,d,F', 2.2),
    'Ga13': ('4s,s,4p,p,3d,d,F', 2.2),
    'Ge4': ('4s,s,4p,p,d,F', 2.1),
    'Ge14': ('4s,s,4p,p,3d,d,F', 2.1),
    'As5': ('4s,s,4p,p,d,F', 2.0),
    'Se6': ('4s,s,4p,p,d,F', 2.1),
    'Br7': ('4s,s,4p,p,d,F', 2.1),
    'Kr8': ('4s,s,4p,p,d,F', 2.1),
    # 37-54:
    'Rb1': ('5s,s,5p', 3.6),
    'Rb9': ('4s,5s,4p,5p,d,d,F', 2.5),
    'Sr2': ('5s,s,5p', 3.3),
    'Sr10': ('4s,5s,4p,5p,4d,d,F', 2.5),
    'Y3': ('5s,s,5p,p,4d,d', 3.1),
    'Y11': ('4s,5s,4p,5p,4d,d,F', 2.5),
    'Zr4': ('5s,s,5p,p,4d,d', 3.0),
    'Zr12': ('4s,5s,4p,5p,4d,d,F', 2.5),
    'Nb5': ('5s,s,5p,p,4d,d', 2.9),
    'Nb13': ('4s,5s,4p,5p,4d,d,F', [2.4, 2.4, 2.5]),
    'Mo6': ('5s,s,5p,p,4d,d', 2.8),
    'Mo14': ('4s,5s,4p,5p,4d,d,F', 2.3),
    'Tc7': ('5s,s,5p,p,4d,d', 2.7),
    'Tc15': ('4s,5s,4p,5p,4d,d,F', 2.3),
    'Ru8': ('5s,s,5p,p,4d,d', 2.6),
    'Ru16': ('4s,5s,4p,5p,4d,d,F', 2.3),
    'Rh9': ('5s,s,5p,p,4d,d', 2.5),
    'Rh17': ('4s,5s,4p,5p,4d,d,F', 2.3),
    'Pd10': ('5s,s,5p,p,4d,d', 2.4),
    'Pd18': ('4s,5s,4p,5p,4d,d,F', 2.3),
    'Ag11': ('5s,s,5p,p,4d,d', 2.4),
    'Ag19': ('4s,5s,4p,5p,4d,d,F', 2.3),
    'Cd12': ('5s,s,5p,p,4d,d', 2.4),
    'Cd20': ('4s,5s,4p,5p,4d,d,F', 2.3),
    'In13': ('5s,s,5p,p,4d,d,F', 2.6),
    'Sn14': ('5s,s,5p,p,4d,d,F', 2.5),
    'Sb15': ('5s,s,5p,p,4d,d,F', 2.5),
    'Te6': ('5s,6s,5p,p,d,d,F', 2.5),
    'I7': ('5s,s,5p,p,d,F', 2.4),
    'Xe8': ('5s,s,5p,p,d,F', 2.3),
    # 55-56:
    'Cs1': ('6s,s,6p,5d', [4.3, 4.6, 4.0]),
    'Cs9': ('5s,6s,5p,6p,5d,0.5d,F', 3.2),
    'Ba2': ('6s,s,6p,5d', 3.9),
    'Ba10': ('5s,6s,5p,6p,5d,d,F', 2.2),
    # 57-71:
    'La11': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.5),
    'Ce12': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.4),
    'Pr13': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.3),
    'Nd14': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.3),
    'Pm15': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.3),
    'Sm16': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Eu17': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Gd18': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Tb19': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Dy20': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.1),
    'Ho21': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Er22': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Tm23': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Yb24': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    'Lu25': ('5s,6s,5p,6p,5d,d,4f,f,G', 2.2),
    # 72-86:
    'Hf4': ('6s,s,6p,p,5d,d', 2.9),
    'Hf12': ('5s,6s,5p,6p,5d,d,F', 2.4),
    'Ta5': ('6s,s,6p,p,5d,d', 2.8),
    'Ta13': ('5s,6s,5p,6p,5d,d,F', 2.4),
    'W6': ('6s,s,6p,p,5d,d', 2.7),
    'W14': ('5s,6s,5p,6p,5d,d,F', 2.4),
    'Re7': ('6s,s,6p,p,5d,d', 2.6),
    'Re15': ('5s,6s,5p,6p,5d,d,F', 2.4),
    'Os8': ('6s,s,6p,p,5d,d', 2.6),
    'Os16': ('5s,6s,5p,6p,5d,d,F', 2.4),
    'Ir9': ('6s,s,6p,p,5d,d', 2.6),
    'Ir17': ('5s,6s,5p,6p,5d,d,F', 2.4),
    'Pt10': ('6s,s,6p,p,5d,d', 2.5),
    'Pt18': ('5s,6s,5p,6p,5d,d,F', 2.3),
    'Au11': ('6s,s,6p,p,5d,d', 2.5),
    'Au19': ('5s,6s,5p,6p,5d,d,F', 2.3),
    'Hg12': ('6s,s,6p,p,5d,d', 2.5),
    'Hg20': ('5s,6s,5p,6p,5d,d,F', 2.3),
    'Tl13': ('6s,s,6p,p,5d,d,F', 2.8),
    'Pb14': ('6s,s,6p,p,5d,d,F', 2.6),
    'Bi5': ('6s,s,6p,p,d,F', 2.8),
    'Bi15': ('6s,s,6p,p,5d,d,F', 2.6),
    'Po6': ('6s,s,6p,p,d,F', 2.7),
    'At7': ('6s,s,6p,p,d,F', 2.6),
    'Rn8': ('6s,s,6p,p,d,F', 2.6),
    # 87-88:
    'Fr1': ('6s,s,6p,5d', 4.5),
    'Fr9': ('6s,7s,6p,7p,6d,d,F', [2.7, 2.5]),
    'Ra2': ('6s,s,6p,5d', 4.5),
    'Ra10': ('6s,7s,6p,7p,6d,d,F', [2.7, 2.5]),
    # 89-102:
    'Ac11': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Th12': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.4),
    'Pa13': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'U14': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Np15': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Pu16': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Am17': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Cm18': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Bk19': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Cf20': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Es21': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Fm22': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'Md23': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5),
    'No24': ('6s,7s,6p,7p,6d,d,5f,f,G', 2.5)}


default = [0,
           1, 2,
           1, 2, 3, 4, 5, 6, 7, 8,
           1, 2, 3, 4, 5, 6, 7, 8,
           1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 3, 4, 5, 6, 7, 8,
           1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 6, 7, 8,
           1, 2, 11,
           12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,
           4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 6, 7, 8,
           9, 10,
           11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]


semicore = [0,
            1, 2,
            3, 4, 3, 4, 5, 6, 7, 8,
            9, 10, 3, 4, 5, 6, 7, 8,
            9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 13, 14, 5, 6, 7, 8,
            9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 13, 14, 15, 6, 7, 8,
            9, 10, 11,
            12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,
            12, 13, 14, 15, 16, 17, 18, 19, 20, 13, 14, 15, 6, 7, 8]


def get_number_of_electrons(symbol, name):
    Z = atomic_numbers[symbol]
    if name == 'default':
        return default[Z]
    return semicore[Z]


class PAWWaves:
    def __init__(self, rgd, l, rcut):
        self.rgd = rgd
        self.l = l
        self.rcut = rcut

        self.n_n = []
        self.e_n = []
        self.f_n = []
        self.phi_ng = []
        self.phit_ng = None
        self.pt_ng = None

    def __len__(self):
        return len(self.n_n)

    def add(self, phi_g, n, e, f):
        self.phi_ng.append(phi_g)
        self.n_n.append(n)
        self.e_n.append(e)
        self.f_n.append(f)

    def pseudize(self, type, nderiv, vtr_g, vr_g, rcmax):
        rgd = self.rgd
        r_g = rgd.r_g
        phi_ng = self.phi_ng = np.array(self.phi_ng)
        N = len(phi_ng)
        phit_ng = self.phit_ng = rgd.empty(N)
        pt_ng = self.pt_ng = rgd.empty(N)
        gc = rgd.ceil(self.rcut)
        gcmax = rgd.ceil(rcmax)

        l = self.l

        dgdr_g = 1 / rgd.dr_g
        d2gdr2_g = rgd.d2gdr2()

        self.nt_g = rgd.zeros()
        for n, phi_g in enumerate(phi_ng):
            if type == 'poly':
                phit_ng[n], c0 = rgd.pseudize(phi_g, gc, self.l, nderiv)
            elif type == 'nc':
                phit_ng[n], c0 = rgd.pseudize_normalized(phi_g, gc, self.l,
                                                         nderiv)
            else:
                1 / 0
            a_g, dadg_g, d2adg2_g = rgd.zeros(3)
            a_g[1:] = self.phit_ng[n, 1:] / r_g[1:]**l
            a_g[0] = c0
            dadg_g[1:-1] = 0.5 * (a_g[2:] - a_g[:-2])
            d2adg2_g[1:-1] = a_g[2:] - 2 * a_g[1:-1] + a_g[:-2]
            q_g = (vtr_g - self.e_n[n] * r_g) * self.phit_ng[n]
            q_g -= 0.5 * r_g**l * (
                (2 * (l + 1) * dgdr_g + r_g * d2gdr2_g) * dadg_g +
                r_g * d2adg2_g * dgdr_g**2)
            q_g[gcmax:] = 0
            rgd.cut(q_g, self.rcut)
            q_g[1:] /= r_g[1:]
            if l == 0:
                q_g[0] = q_g[1]
            pt_ng[n] = q_g

            self.nt_g += self.f_n[n] / 4 / pi * phit_ng[n]**2

        self.dS_nn = (rgd.integrate(phi_ng[:, None] * phi_ng) -
                      rgd.integrate(phit_ng[:, None] * phit_ng)) / (4 * pi)
        self.Q = np.dot(self.f_n, self.dS_nn.diagonal())
        A_nn = rgd.integrate(phit_ng[:, None] * pt_ng) / (4 * pi)
        self.dH_nn = np.dot(self.dS_nn, np.diag(self.e_n)) - A_nn
        self.dH_nn *= 0.5
        self.dH_nn += self.dH_nn.T.copy()
        pt_ng[:] = np.dot(np.linalg.inv(A_nn.T), pt_ng)

    def construct_projectors(self, vtr_g, rcmax):
        pass

    def calculate_kinetic_energy_correction(self, vr_g, vtr_g):
        if len(self) == 0:
            return
        self.dekin_nn = (self.rgd.integrate(self.phit_ng[:, None] *
                                            self.phit_ng *
                                            vtr_g, -1) / (4 * pi) -
                         self.rgd.integrate(self.phi_ng[:, None] *
                                            self.phi_ng *
                                            vr_g, -1) / (4 * pi) +
                         self.dH_nn)


class PAWSetupGenerator:
    def __init__(self, aea, projectors,
                 scalar_relativistic=False,
                 core_hole=None,
                 fd=None, yukawa_gamma=0.0, omega=None):
        """fd: stream
            Text output.

           yukawa_gamma: separation parameter for RSF"""

        self.aea = aea

        self.fd = fd or sys.stdout
        self.yukawa_gamma = yukawa_gamma
        self.exxcc_w = {}
        self.exxcv_wii = {}
        self.omega = omega

        if core_hole:
            state, occ = core_hole.split(',')
            n = int(state[0])
            l = 'spdf'.find(state[1])
            occ = float(occ)
            aea.add(n, l, -occ)
            self.core_hole = (n, l, occ)
        else:
            self.core_hole = None

        if projectors[-1].isupper():
            assert projectors[-2] == ',', projectors
            self.l0 = 'SPDFG'.find(projectors[-1])
            projectors = projectors[:-2]
        else:
            self.l0 = None

        self.lmax = -1
        self.states = {}
        for s in projectors.split(','):
            l = 'spdf'.find(s[-1])
            if len(s) == 1:
                n = None
            elif '.' in s:
                n = float(s[:-1])
            else:
                n = int(s[:-1])
            if l in self.states:
                self.states[l].append(n)
            else:
                self.states[l] = [n]
            if l > self.lmax:
                self.lmax = l

        # Add empty bound states:
        for l, nn in self.states.items():
            for n in nn:
                if (isinstance(n, int) and
                    (l not in aea.f_lsn or n - l > len(aea.f_lsn[l][0]))):
                    aea.add(n, l, 0)

        aea.initialize()
        aea.run()
        aea.scalar_relativistic = scalar_relativistic
        aea.refine()

        self.rgd = aea.rgd

        self.vtr_g = None

        self.basis = None

        self.log('\nGenerating PAW', aea.xc.name, 'setup for', aea.symbol)

    def construct_shape_function(self, alpha=None, rc=None, eps=1e-10):
        """Build shape-function for compensation charge."""

        self.alpha = alpha

        if self.alpha is None:
            if not isinstance(rc, (float, int)):
                rc = min(rc)
            rc = 1.5 * rc

            def spillage(alpha):
                """Fraction of gaussian charge outside rc."""
                x = alpha * rc**2
                return 1 - erf(sqrt(x)) + 2 * sqrt(x / pi) * exp(-x)

            def f(alpha):
                return log(spillage(alpha)) - log(eps)

            self.alpha = fsolve(f, 7.0)[0]

            self.alpha = round(self.alpha, 1)
        elif alpha < 0:
            rc = min(rc)
            self.log('Shape function: (sin(pi*r/rc)/r)^2, rc={rc:.2f} Bohr'
                     .format(rc=rc))
            self.ghat_g = np.sinc(self.rgd.r_g / rc)**2 * (pi / 2 / rc**3)
            self.ghat_g[self.rgd.ceil(rc):] = 0.0
            self.alpha = -rc
            return

        self.log('Shape function: exp(-alpha*r^2), alpha=%.1f Bohr^-2' %
                 self.alpha)

        self.ghat_g = (np.exp(-self.alpha * self.rgd.r_g**2) *
                       (self.alpha / pi)**1.5)

    def log(self, *args, **kwargs):
        print(file=self.fd, *args, **kwargs)

    def calculate_core_density(self):
        self.nc_g = self.rgd.zeros()
        self.tauc_g = self.rgd.zeros()
        self.ncore = 0
        self.nvalence = 0
        self.ekincore = 0.0
        for l, ch in enumerate(self.aea.channels):
            for n, f in enumerate(ch.f_n):
                if (l <= self.lmax and
                    any(n + l + 1 == nn
                        for nn in self.states[l]
                        if isinstance(nn, int))):
                    self.nvalence += f
                else:
                    self.nc_g += f * ch.calculate_density(n)
                    self.tauc_g += f * ch.calculate_kinetic_energy_density(n)
                    self.ncore += f
                    self.ekincore += f * ch.e_n[n]

        self.ekincore -= self.rgd.integrate(self.nc_g * self.aea.vr_sg[0], -1)

        self.log('Core electrons:', self.ncore)
        self.log('Valence electrons:', self.nvalence)

    def find_local_potential(self, r0, P, dv0=None):
        self.r0 = r0
        self.nderiv0 = P
        if self.l0 is None:
            self.find_polynomial_potential(r0, P, dv0)
        else:
            self.match_local_potential(r0, P)

    def find_polynomial_potential(self, r0, P, dv0):
        self.log('Constructing smooth local potential for r < %.3f' % r0)
        g0 = self.rgd.ceil(r0)
        self.vtr_g = self.rgd.pseudize(self.aea.vr_sg[0], g0, 1, P)[0]
        if dv0:
            x = self.rgd.r_g[:g0] / r0
            dv_g = dv0 * (1 - 2 * x**2 + x**4)
            self.vtr_g[:g0] += x * r0 * dv_g

    def match_local_potential(self, r0, P):
        l0 = self.l0
        self.log('Local potential matching %s-scattering at e=0.0 eV' %
                 'spdfg'[l0] + ' and r=%.2f Bohr' % r0)

        g0 = self.rgd.ceil(r0)
        gc = g0 + 20

        e0 = 0.0

        ch = Channel(l0)
        phi_g = self.rgd.zeros()
        a = ch.integrate_outwards(phi_g, self.rgd, self.aea.vr_sg[0], gc, e0,
                                  self.aea.scalar_relativistic, self.aea.Z)[1]
        phi_g[1:gc] /= self.rgd.r_g[1:gc]
        phi_g[0] = a

        phit_g, c = self.rgd.pseudize(phi_g, g0, l=l0, points=P)

        dgdr_g = 1 / self.rgd.dr_g
        d2gdr2_g = self.rgd.d2gdr2()
        a_g = phit_g.copy()
        a_g[1:] /= self.rgd.r_g[1:]**l0
        a_g[0] = c
        dadg_g = self.rgd.zeros()
        d2adg2_g = self.rgd.zeros()
        dadg_g[1:-1] = 0.5 * (a_g[2:] - a_g[:-2])
        d2adg2_g[1:-1] = a_g[2:] - 2 * a_g[1:-1] + a_g[:-2]
        q_g = (((l0 + 1) * dgdr_g + 0.5 * self.rgd.r_g * d2gdr2_g) * dadg_g +
               0.5 * self.rgd.r_g * d2adg2_g * dgdr_g**2)
        q_g[:g0] /= a_g[:g0]
        q_g += e0 * self.rgd.r_g
        q_g[0] = 0.0

        self.vtr_g = self.aea.vr_sg[0].copy()
        self.vtr_g[0] = 0.0
        self.vtr_g[1:g0] = q_g[1:g0]

    def add_waves(self, rc):
        if isinstance(rc, float):
            radii = [rc]
        else:
            radii = list(rc)

        if self.lmax >= 0:
            radii += [radii[-1]] * (self.lmax + 1 - len(radii))
        del radii[self.lmax + 1:]  # remove unused radii

        self.rcmax = max(radii)

        self.waves_l = []
        for l in range(self.lmax + 1):
            rcut = radii[l]
            waves = PAWWaves(self.rgd, l, rcut)
            e = -1.0
            for n in self.states[l]:
                if isinstance(n, int):
                    # Bound state:
                    ch = self.aea.channels[l]
                    e = ch.e_n[n - l - 1]
                    f = ch.f_n[n - l - 1]
                    phi_g = ch.phi_ng[n - l - 1]
                else:
                    if n is None:
                        e += 1.0
                    else:
                        e = n
                    n = -1
                    f = 0.0
                    phi_g = self.rgd.zeros()
                    gc = self.rgd.round(2.5 * self.rcmax)
                    ch = Channel(l)
                    a = ch.integrate_outwards(phi_g, self.rgd,
                                              self.aea.vr_sg[0], gc, e,
                                              self.aea.scalar_relativistic,
                                              self.aea.Z)[1]
                    phi_g[1:gc + 1] /= self.rgd.r_g[1:gc + 1]
                    phi_g[0] = a
                    phi_g /= (self.rgd.integrate(phi_g**2) / (4 * pi))**0.5

                waves.add(phi_g, n, e, f)
            self.waves_l.append(waves)

    def pseudize(self, type='poly', nderiv=6, rcore=None):
        self.Q = -self.aea.Z + self.ncore

        self.nt_g = self.rgd.zeros()
        for waves in self.waves_l:
            waves.pseudize(type, nderiv, self.vtr_g, self.aea.vr_sg[0],
                           2.0 * self.rcmax)
            self.nt_g += waves.nt_g
            self.Q += waves.Q

        self.construct_pseudo_core_density(rcore)
        self.calculate_potentials()
        self.summarize()

    def construct_pseudo_core_density(self, rcore):
        if rcore is None:
            rcore = self.rcmax * 0.8
        else:
            assert abs(rcore) <= self.rcmax

        if self.ncore == 0:
            self.nct_g = self.rgd.zeros()
            self.tauct_g = self.rgd.zeros()
        elif rcore > 0.0:
            # Make sure pseudo density is monotonically decreasing:
            while True:
                gcore = self.rgd.round(rcore)
                self.nct_g = self.rgd.pseudize(self.nc_g, gcore)[0]
                nt_g = self.nt_g + self.nct_g
                dntdr_g = self.rgd.derivative(nt_g)[:gcore]
                if dntdr_g.max() < 0.0:
                    break
                rcore -= 0.01

            rcore *= 1.2
            gcore = self.rgd.round(rcore)
            self.nct_g = self.rgd.pseudize(self.nc_g, gcore)[0]
            nt_g = self.nt_g + self.nct_g

            self.log('Constructing smooth pseudo core density for r < %.3f' %
                     rcore)
            self.nt_g = nt_g

            self.tauct_g = self.rgd.pseudize(self.tauc_g, gcore)[0]
        else:
            rcore *= -1
            gcore = self.rgd.round(rcore)
            nt_g = self.rgd.pseudize(self.aea.n_sg[0], gcore)[0]
            self.nct_g = nt_g - self.nt_g
            self.nt_g = nt_g

            self.log('Constructing NLCC-style smooth pseudo core density for '
                     'r < %.3f' % rcore)

            self.tauct_g = self.rgd.pseudize(self.tauc_g, gcore)[0]

        self.npseudocore = self.rgd.integrate(self.nct_g)
        self.log('Pseudo core electrons: %.6f' % self.npseudocore)
        self.Q -= self.npseudocore

    def calculate_potentials(self):
        self.rhot_g = self.nt_g + self.Q * self.ghat_g
        self.vHtr_g = self.rgd.poisson(self.rhot_g)

        self.vxct_g = self.rgd.zeros()
        nt_sg = self.nt_g.reshape((1, -1))
        self.exct = self.aea.xc.calculate_spherical(
            self.rgd, nt_sg, self.vxct_g.reshape((1, -1)))

        self.v0r_g = self.vtr_g - self.vHtr_g - self.vxct_g * self.rgd.r_g
        self.v0r_g[self.rgd.round(self.rcmax):] = 0.0

    def summarize(self):
        self.log('\nProjectors:')
        self.log(' state  occ         energy             norm        rcut')
        self.log(' nl            [Hartree]  [eV]      [electrons]   [Bohr]')
        self.log('----------------------------------------------------------')
        for l, waves in enumerate(self.waves_l):
            for n, e, f, ds in zip(waves.n_n, waves.e_n, waves.f_n,
                                   waves.dS_nn.diagonal()):
                if n == -1:
                    self.log('  %s         %10.6f %10.5f   %19.2f' %
                             ('spdf'[l], e, e * Hartree, waves.rcut))
                else:
                    self.log(
                        ' %d%s   %5.2f %10.6f %10.5f      %5.3f  %9.2f' %
                        (n, 'spdf'[l], f, e, e * Hartree, 1 - ds,
                         waves.rcut))
        self.log()

    def construct_projectors(self, rcore):
        for waves in self.waves_l:
            waves.construct_projectors(self.vtr_g, 2.45 * self.rcmax)
            waves.calculate_kinetic_energy_correction(self.aea.vr_sg[0],
                                                      self.vtr_g)

    def check_all(self):
        self.log(('Checking eigenvalues of %s pseudo atom using ' +
                  'a Gaussian basis set:') % self.aea.symbol)
        self.log('                 AE [eV]        PS [eV]      error [eV]')

        ok = True

        for l in range(4):
            try:
                e_b = self.check(l)
            except RuntimeError:
                self.log('Singular overlap matrix!')
                ok = False
                continue

            n0 = self.number_of_core_states(l)

            if l < len(self.aea.channels):
                e0_b = self.aea.channels[l].e_n
                extra = 6
                nae = len(self.aea.channels[l].f_n)
                for n in range(1 + l, nae + 1 + l + extra):
                    if n - 1 - l < nae:
                        f = self.aea.channels[l].f_n[n - 1 - l]
                        self.log('%2d%s  %2d' % (n, 'spdf'[l], f), end='')
                    else:
                        self.log('       ', end='')
                    self.log('  %15.3f' % (e0_b[n - 1 - l] * Hartree), end='')
                    if n - 1 - l - n0 >= 0:
                        self.log('%15.3f' * 2 %
                                 (e_b[n - 1 - l - n0] * Hartree,
                                  (e_b[n - 1 - l - n0] - e0_b[n - 1 - l]) *
                                  Hartree))
                    else:
                        self.log()

                errors = abs(e_b[:nae - n0] - e0_b[n0:nae])
                if (errors > 2e-3).any():
                    self.log('Error in bound %s-states!' % 'spdf'[l])
                    ok = False
                errors = abs(e_b[nae - n0:nae - n0 + extra] -
                             e0_b[nae:nae + extra])
                if (not self.aea.scalar_relativistic and
                    (errors > 2e-2).any()):
                    self.log('Error in %s-states!' % 'spdf'[l])
                    ok = False

        return ok

    def number_of_core_states(self, l):
        n0 = 0
        if l < len(self.waves_l):
            waves = self.waves_l[l]
            if len(waves) > 0:
                n0 = waves.n_n[0] - l - 1
                if n0 < 0 and l < len(self.aea.channels):
                    n0 = (self.aea.channels[l].f_n > 0).sum()
        elif l < len(self.aea.channels):
            n0 = (self.aea.channels[l].f_n > 0).sum()
        return n0

    def check(self, l):
        basis = self.aea.channels[0].basis
        eps = basis.eps
        alpha_B = basis.alpha_B

        basis = GaussianBasis(l, alpha_B, self.rgd, eps)
        H_bb = basis.calculate_potential_matrix(self.vtr_g)
        H_bb += basis.T_bb
        S_bb = np.eye(len(basis))

        if l < len(self.waves_l):
            waves = self.waves_l[l]
            if len(waves) > 0:
                P_bn = self.rgd.integrate(basis.basis_bg[:, None] *
                                          waves.pt_ng) / (4 * pi)
                H_bb += np.dot(np.dot(P_bn, waves.dH_nn), P_bn.T)
                S_bb += np.dot(np.dot(P_bn, waves.dS_nn), P_bn.T)

        e_b, _ = eigh(H_bb, S_bb)
        return e_b

    def test_convergence(self):
        rgd = self.rgd
        r_g = rgd.r_g
        G_k, nt_k = self.rgd.fft(self.nt_g * r_g)
        rhot_k = self.rgd.fft(self.rhot_g * r_g)[1]
        ghat_k = self.rgd.fft(self.ghat_g * r_g)[1]
        vt_k = self.rgd.fft(self.vtr_g)[1]
        phi_k = self.rgd.fft(self.waves_l[0].phit_ng[0] * r_g)[1]
        eee_k = 0.5 * nt_k**2 * (4 * pi)**2 / (2 * pi)**3
        ecc_k = 0.5 * rhot_k**2 * (4 * pi)**2 / (2 * pi)**3
        egg_k = 0.5 * ghat_k**2 * (4 * pi)**2 / (2 * pi)**3
        ekin_k = 0.5 * phi_k**2 * G_k**4 / (2 * pi)**3
        evt_k = nt_k * vt_k * G_k**2 * 4 * pi / (2 * pi)**3

        eee = 0.5 * rgd.integrate(self.nt_g * rgd.poisson(self.nt_g), -1)
        ecc = 0.5 * rgd.integrate(self.rhot_g * self.vHtr_g, -1)
        egg = 0.5 * rgd.integrate(self.ghat_g * rgd.poisson(self.ghat_g), -1)
        ekin = self.aea.ekin - self.ekincore - self.waves_l[0].dekin_nn[0, 0]
        evt = rgd.integrate(self.nt_g * self.vtr_g, -1)

        import matplotlib.pyplot as plt

        errors = 10.0**np.arange(-4, 0) / Hartree
        self.log('\nConvergence of energy:')
        self.log('plane-wave cutoff (wave-length) [ev (Bohr)]\n  ', end='')
        for de in errors:
            self.log('%14.4f' % (de * Hartree), end='')
        for label, e_k, e0 in [
            ('e-e', eee_k, eee),
            ('c-c', ecc_k, ecc),
            ('g-g', egg_k, egg),
            ('kin', ekin_k, ekin),
            ('vt', evt_k, evt)]:
            self.log('\n%3s: ' % label, end='')
            e_k = (np.add.accumulate(e_k) - 0.5 * e_k[0] - 0.5 * e_k) * G_k[1]
            k = len(e_k) - 1
            for de in errors:
                while abs(e_k[k] - e_k[-1]) < de:
                    k -= 1
                G = k * G_k[1]
                ecut = 0.5 * G**2
                h = pi / G
                self.log(' %6.1f (%4.2f)' % (ecut * Hartree, h), end='')
            plt.semilogy(G_k, abs(e_k - e_k[-1]) * Hartree, label=label)
        self.log()
        plt.axis(xmax=20)
        plt.xlabel('G')
        plt.ylabel('[eV]')
        plt.legend()
        plt.show()

    def plot(self):
        import matplotlib.pyplot as plt
        r_g = self.rgd.r_g

        plt.figure()
        plt.plot(r_g, self.vxct_g, label='xc')
        plt.plot(r_g[1:], self.v0r_g[1:] / r_g[1:], label='0')
        plt.plot(r_g[1:], self.vHtr_g[1:] / r_g[1:], label='H')
        plt.plot(r_g[1:], self.vtr_g[1:] / r_g[1:], label='ps')
        plt.plot(r_g[1:], self.aea.vr_sg[0, 1:] / r_g[1:], label='ae')
        plt.axis(xmin=0,
                 xmax=2 * self.rcmax,
                 ymin=self.vtr_g[1] / r_g[1],
                 ymax=max(0, (self.v0r_g[1:] / r_g[1:]).max()))
        plt.xlabel('radius [Bohr]')
        plt.ylabel('potential [Ha]')
        plt.legend()

        plt.figure()
        i = 0
        for l, waves in enumerate(self.waves_l):
            for n, e, phi_g, phit_g in zip(waves.n_n, waves.e_n,
                                           waves.phi_ng, waves.phit_ng):
                if n == -1:
                    gc = self.rgd.ceil(waves.rcut)
                    name = '*%s (%.2f Ha)' % ('spdf'[l], e)
                else:
                    gc = len(self.rgd)
                    name = '%d%s (%.2f Ha)' % (n, 'spdf'[l], e)
                plt.plot(r_g[:gc], (phi_g * r_g)[:gc], color=colors[i],
                         label=name)
                plt.plot(r_g[:gc], (phit_g * r_g)[:gc], '--', color=colors[i])
                i += 1
        plt.axis(xmin=0, xmax=3 * self.rcmax)
        plt.xlabel('radius [Bohr]')
        plt.ylabel(r'$r\phi_{n\ell}(r)$')
        plt.legend()

        plt.figure()
        i = 0
        for l, waves in enumerate(self.waves_l):
            for n, e, pt_g in zip(waves.n_n, waves.e_n, waves.pt_ng):
                if n == -1:
                    name = '*%s (%.2f Ha)' % ('spdf'[l], e)
                else:
                    name = '%d%s (%.2f Ha)' % (n, 'spdf'[l], e)
                plt.plot(r_g, pt_g * r_g, color=colors[i], label=name)
                i += 1
        plt.axis(xmin=0, xmax=self.rcmax)
        plt.legend()

    def create_basis_set(self, tailnorm=0.0005, scale=200.0, splitnorm=0.16,
                         tag=None, ri=None):
        rgd = self.rgd
        name = 'dzp' if not tag else f'{tag}.dzp'
        self.basis = Basis(self.aea.symbol, name, readxml=False, rgd=rgd)

        # We print text to sdtout and put it in the basis-set file
        txt = 'Basis functions:\n'

        # Bound states:
        for l, waves in enumerate(self.waves_l):
            for i, n in enumerate(waves.n_n):
                if n > 0:
                    tn = tailnorm
                    if waves.f_n[i] == 0:
                        tn = min(0.05, tn * 20)  # no need for long tail
                    phit_g, ronset, rc, de = self.create_basis_function(
                        l, i, tn, scale)
                    bf = BasisFunction(n, l, rc, phit_g, 'bound state')
                    self.basis.append(bf)

                    txt += '%d%s bound state:\n' % (n, 'spdf'[l])
                    txt += ('  cutoff: %.3f to %.3f Bohr (tail-norm=%f)\n' %
                            (ronset, rc, tn))
                    txt += '  eigenvalue shift: %.3f eV\n' % (de * Hartree)

        # Split valence:
        for l, waves in enumerate(self.waves_l):
            # Find the largest n that is occupied:
            n0 = None
            for f, n in zip(waves.f_n, waves.n_n):
                if n > 0 and f > 0:
                    n0 = n
            if n0 is None:
                continue

            for bf in self.basis.bf_j:
                if bf.l == l and bf.n == n0:
                    break

            # Radius and l-value used for polarization function below:
            rcpol = bf.rc
            lpol = l + 1

            phit_g = bf.phit_g

            # Find cutoff radius:
            n_g = np.add.accumulate(phit_g**2 * rgd.r_g**2 * rgd.dr_g)
            norm = n_g[-1]
            gc = (norm - n_g > splitnorm * norm).sum()
            rc = rgd.r_g[gc]

            phit2_g = rgd.pseudize(phit_g, gc, l, 2)[0]  # "split valence"
            bf = BasisFunction(n, l, rc, phit_g - phit2_g, 'split valence')
            self.basis.append(bf)

            txt += '%d%s split valence:\n' % (n0, 'spdf'[l])
            txt += '  cutoff: %.3f Bohr (tail-norm=%f)\n' % (rc, splitnorm)

        # Polarization:
        if lpol < 4:
            gcpol = rgd.round(rcpol)
            alpha = 1 / (0.25 * rcpol)**2

            # Gaussian that is continuous and has a continuous
            # derivative at rcpol:
            phit_g = np.exp(-alpha * rgd.r_g**2) * rgd.r_g**lpol
            phit_g -= rgd.pseudize(phit_g, gcpol, lpol, 2)[0]
            phit_g[gcpol:] = 0.0

            bf = BasisFunction(None, lpol, rcpol, phit_g, 'polarization')
            self.basis.append(bf)
            txt += f'l={lpol} polarization functions:\n'
            txt += f'  cutoff: {rcpol:.3f} Bohr '
            txt += f'(r^{lpol} exp(-{alpha:.3f}*r^2))\n'

        self.log(txt)

        # Write basis-set file:
        self.basis.generatordata = txt
        self.basis.generatorattrs.update(dict(tailnorm=tailnorm,
                                              scale=scale,
                                              splitnorm=splitnorm))

        if ri:
            generate_ri_basis(self.basis, ri)

        return self.basis

    def create_basis_function(self, l, n, tailnorm, scale):
        rgd = self.rgd
        waves = self.waves_l[l]

        # Find cutoff radii:
        n_g = np.add.accumulate(waves.phit_ng[n]**2 * rgd.r_g**2 * rgd.dr_g)
        norm = n_g[-1]
        g2 = (norm - n_g > tailnorm * norm).sum()
        r2 = rgd.r_g[g2]
        r1 = max(0.6 * r2, waves.rcut)
        g1 = rgd.ceil(r1)
        # Set up confining potential:
        r = rgd.r_g[g1:g2]
        vtr_g = self.vtr_g.copy()
        vtr_g[g1:g2] += scale * np.exp((r2 - r1) / (r1 - r)) / (r - r2)**2
        vtr_g[g2:] = np.inf

        # Nonlocal PAW stuff:
        pt_ng = waves.pt_ng
        dH_nn = waves.dH_nn
        dS_nn = waves.dS_nn
        N = len(pt_ng)

        u_g = rgd.zeros()
        u_ng = rgd.zeros(N)
        duodr_n = np.empty(N)
        a_n = np.empty(N)

        e = waves.e_n[n]
        e0 = e
        ch = Channel(l)
        while True:
            duodr, a = ch.integrate_outwards(u_g, rgd, vtr_g, g1, e)

            for n in range(N):
                duodr_n[n], a_n[n] = ch.integrate_outwards(u_ng[n], rgd,
                                                           vtr_g, g1, e,
                                                           pt_g=pt_ng[n])

            A_nn = (dH_nn - e * dS_nn) / (4 * pi)
            B_nn = rgd.integrate(pt_ng[:, None] * u_ng, -1)
            c_n = rgd.integrate(pt_ng * u_g, -1)
            d_n = np.linalg.solve(np.dot(A_nn, B_nn) + np.eye(N),
                                  np.dot(A_nn, c_n))
            u_g[:g1 + 1] -= np.dot(d_n, u_ng[:, :g1 + 1])
            a -= np.dot(d_n, a_n)
            duodr -= np.dot(duodr_n, d_n)
            uo = u_g[g1]

            duidr = ch.integrate_inwards(u_g, rgd, vtr_g, g1, e, gmax=g2)
            ui = u_g[g1]
            A = duodr / uo - duidr / ui
            u_g[g1:] *= uo / ui
            x = (norm / rgd.integrate(u_g**2, -2) * (4 * pi))**0.5
            u_g *= x
            a *= x

            if abs(A) < 1e-5:
                break

            e += 0.5 * A * u_g[g1]**2

        u_g[1:] /= rgd.r_g[1:]
        u_g[0] = a * 0.0**l
        return u_g, r1, r2, e - e0

    def logarithmic_derivative(self, l, energies, rcut):
        rgd = self.rgd
        ch = Channel(l)
        gcut = rgd.round(rcut)

        N = 0
        if l < len(self.waves_l):
            # Nonlocal PAW stuff:
            waves = self.waves_l[l]
            if len(waves) > 0:
                pt_ng = waves.pt_ng
                dH_nn = waves.dH_nn
                dS_nn = waves.dS_nn
                N = len(pt_ng)

        u_g = rgd.zeros()
        u_ng = rgd.zeros(N)
        dudr_n = np.empty(N)

        logderivs = []
        d0 = 42.0
        offset = 0
        for e in energies:
            dudr = ch.integrate_outwards(u_g, rgd, self.vtr_g, gcut, e)[0]
            u = u_g[gcut]

            if N:
                for n in range(N):
                    dudr_n[n] = ch.integrate_outwards(u_ng[n], rgd,
                                                      self.vtr_g, gcut, e,
                                                      pt_g=pt_ng[n])[0]

                A_nn = (dH_nn - e * dS_nn) / (4 * pi)
                B_nn = rgd.integrate(pt_ng[:, None] * u_ng, -1)
                c_n = rgd.integrate(pt_ng * u_g, -1)
                d_n = np.linalg.solve(np.dot(A_nn, B_nn) + np.eye(N),
                                      np.dot(A_nn, c_n))
                u -= np.dot(u_ng[:, gcut], d_n)
                dudr -= np.dot(dudr_n, d_n)

            d1 = np.arctan(dudr / u) / pi + offset
            if d1 > d0:
                offset -= 1
                d1 -= 1
            logderivs.append(d1)
            d0 = d1

        return np.array(logderivs)

    def make_paw_setup(self, tag=None):
        aea = self.aea

        from gpaw.setup_data import SetupData
        setup = SetupData(aea.symbol, aea.xc.name, tag, readxml=False)

        setup.id_j = []

        J = []  # new reordered j and i indices
        I = []

        # Bound states:
        j = 0
        i = 0
        for l, waves in enumerate(self.waves_l):
            for n, f, e, phi_g, phit_g, pt_g in zip(waves.n_n, waves.f_n,
                                                    waves.e_n, waves.phi_ng,
                                                    waves.phit_ng,
                                                    waves.pt_ng):
                if n != -1:
                    setup.append(n, l, f, e, waves.rcut, phi_g, phit_g, pt_g)
                    id = '%d%s' % (n, 'spdf'[l])
                    setup.id_j.append(id)
                    J.append(j)
                    I.extend(range(i, i + 2 * l + 1))
                j += 1
                i += 2 * l + 1

        # Excited states:
        j = 0
        i = 0
        for l, waves in enumerate(self.waves_l):
            ne = 0
            for n, f, e, phi_g, phit_g, pt_g in zip(waves.n_n, waves.f_n,
                                                    waves.e_n, waves.phi_ng,
                                                    waves.phit_ng,
                                                    waves.pt_ng):
                if n == -1:
                    setup.append(n, l, f, e, waves.rcut, phi_g, phit_g, pt_g)
                    ne += 1
                    id = '%s%d' % ('spdf'[l], ne)
                    setup.id_j.append(id)
                    J.append(j)
                    I.extend(range(i, i + 2 * l + 1))
                j += 1
                i += 2 * l + 1

        nj = sum(len(waves) for waves in self.waves_l)
        e_kin_jj = np.zeros((nj, nj))
        j1 = 0
        for waves in self.waves_l:
            j2 = j1 + len(waves)
            e_kin_jj[j1:j2, j1:j2] = waves.dekin_nn
            j1 = j2
        setup.e_kin_jj = e_kin_jj[J][:, J].copy()

        setup.nc_g = self.nc_g * sqrt(4 * pi)
        setup.nct_g = self.nct_g * sqrt(4 * pi)
        setup.e_kinetic_core = self.ekincore
        setup.vbar_g = self.v0r_g * sqrt(4 * pi)
        setup.vbar_g[1:] /= self.rgd.r_g[1:]
        setup.vbar_g[0] = setup.vbar_g[1]
        setup.vt_g = self.vtr_g * sqrt(4 * pi)
        setup.vt_g[1:] /= self.rgd.r_g[1:]
        setup.vt_g[0] = setup.vt_g[1]
        setup.Z = aea.Z
        setup.Nc = self.ncore
        setup.Nv = self.nvalence
        setup.e_kinetic = aea.ekin
        setup.e_xc = aea.exc
        setup.e_electrostatic = aea.eH + aea.eZ
        setup.e_total = aea.exc + aea.ekin + aea.eH + aea.eZ
        setup.rgd = self.rgd
        if self.alpha > 0:
            setup.shape_function = {'type': 'gauss',
                                    'rc': 1 / sqrt(self.alpha)}
        else:
            setup.shape_function = {'type': 'sinc',
                                    'rc': -self.alpha}

        self.calculate_exx_integrals()
        setup.ExxC = self.exxcc
        setup.ExxC_w = self.exxcc_w  # erfc screened core contributions
        setup.X_p = pack2(self.exxcv_ii[I][:, I])
        setup.X_wp = {omega: pack2(self.exxcv_wii[omega][I][:, I])
                      for omega in self.exxcv_wii}

        if self.yukawa_gamma > 0.0:
            self.calculate_yukawa_integrals()
            setup.X_pg = pack2(self.exxgcv_ii[I][:, I])

        setup.tauc_g = self.tauc_g * (4 * pi)**0.5
        setup.tauct_g = self.tauct_g * (4 * pi)**0.5

        if self.aea.scalar_relativistic:
            reltype = 'scalar-relativistic'
        else:
            reltype = 'non-relativistic'
        attrs = [('type', reltype),
                 ('version', 3),
                 ('name', 'gpaw-%s' % version)]
        setup.generatorattrs = attrs
        setup.version = '0.7'

        setup.l0 = self.l0
        setup.e0 = 0.0
        setup.r0 = self.r0
        setup.nderiv0 = self.nderiv0

        setup.basis = self.basis

        if self.core_hole:
            n, l, occ = self.core_hole
            phi_g = self.aea.channels[l].phi_ng[n - l - 1]
            setup.ncorehole = n
            setup.lcorehole = l
            setup.fcorehole = occ
            setup.phicorehole_g = phi_g
            setup.has_corehole = True
            ch = self.aea.channels[l]
            eig = ch.e_n[n - 1 - l]
            setup.core_hole_e = eig
            setup.core_hole_e_kin = eig - self.rgd.integrate(
                ch.calculate_density(n - 1 - l) * self.aea.vr_sg[0], -1)
        return setup

    def find_core_states(self):
        # Find core states:
        core = []
        lmax = 0
        for l, ch in enumerate(self.aea.channels):
            for n, phi_g in enumerate(ch.phi_ng):
                if (l >= len(self.waves_l) or
                    (l < len(self.waves_l) and
                     n + l + 1 not in self.waves_l[l].n_n)):
                    core.append((l, phi_g))
                    if l > lmax:
                        lmax = l

        lmax = max(lmax, len(self.waves_l) - 1)
        G_LLL = gaunt(lmax)

        return lmax, core, G_LLL

    def core_core_exchange(self, interaction):
        # Calculate core contribution to EXX energy per interaction kernel
        (lmax, core, G_LLL) = self.find_core_states()
        E = 0.0
        j1 = 0
        for l1, phi1_g in core:
            f = 1.0
            for l2, phi2_g in core[j1:]:
                n_g = phi1_g * phi2_g
                for l in range((l1 + l2) % 2, l1 + l2 + 1, 2):
                    G = (G_LLL[l1**2:(l1 + 1)**2,
                               l2**2:(l2 + 1)**2,
                               l**2:(l + 1)**2]**2).sum()
                    vr_g = interaction(n_g, l)
                    e = f * self.rgd.integrate(vr_g * n_g, -1) / 4 / pi
                    E -= e * G
                f = 2.0
            j1 += 1
        return E

    def calculate_exx_integrals(self):

        # Calculate core-valence contribution to EXX energy:
        ni = sum(len(waves) * (2 * l + 1)
                 for l, waves in enumerate(self.waves_l))

        self.exxcc = self.core_core_exchange(self.rgd.poisson)
        self.log('EXX (core-core):', self.exxcc, 'Hartree')

        if self.omega is not None:
            hse = RadialHSE(self.rgd, self.omega)

            self.exxcc_w = {self.omega:
                            self.core_core_exchange(hse.screened_coulomb)}
            self.log(f'EXX omega={self.omega} (core-core):',
                     self.exxcc_w[self.omega], 'Hartree')

            X_ii = self.calculate_exx_cv_integrals(ni, hse.screened_coulomb)
            self.exxcv_wii = {self.omega: X_ii}

        self.exxcv_ii = self.calculate_exx_cv_integrals(ni, self.rgd.poisson)

    def calculate_yukawa_integrals(self):
        """Wrapper to calculate the rsf core-valence contribution."""
        ni = sum(len(waves) * (2 * l + 1)
                 for l, waves in enumerate(self.waves_l))

        def interaction(n_g, l):
            return self.rgd.yukawa(n_g, l, self.yukawa_gamma)

        self.exxgcv_ii = self.calculate_exx_cv_integrals(ni, interaction)

    def calculate_exx_cv_integrals(self, ni, interaction):
        (lmax, core, G_LLL) = self.find_core_states()
        cv_ii = np.zeros((ni, ni))

        i1 = 0
        for l1, waves1 in enumerate(self.waves_l):
            for phi1_g in waves1.phi_ng:
                i2 = 0
                for l2, waves2 in enumerate(self.waves_l):
                    for phi2_g in waves2.phi_ng:
                        X_mm = cv_ii[i1:i1 + 2 * l1 + 1,
                                     i2:i2 + 2 * l2 + 1]
                        if (l1 + l2) % 2 == 0:
                            for lc, phi_g in core:
                                n_g = phi1_g * phi_g
                                for l in range((l1 + lc) % 2,
                                               max(l1, l2) + lc + 1, 2):
                                    n2c = phi2_g * phi_g
                                    vr_g = interaction(n2c, l)
                                    e = (self.rgd.integrate(vr_g * n_g, -1) /
                                         (4 * pi))
                                    for mc in range(2 * lc + 1):
                                        for m in range(2 * l + 1):
                                            G_L = G_LLL[:,
                                                        lc**2 + mc,
                                                        l**2 + m]
                                            X_mm += np.outer(
                                                G_L[l1**2:(l1 + 1)**2],
                                                G_L[l2**2:(l2 + 1)**2]) * e
                        i2 += 2 * l2 + 1
                i1 += 2 * l1 + 1
        return cv_ii


def get_parameters(symbol, args):
    Z = atomic_numbers.get(symbol)
    if Z is None:
        Z = float(symbol)
        symbol = chemical_symbols[int(round(Z))]

    if args.electrons:
        par = parameters[symbol + str(args.electrons)]
    else:
        par = parameters[symbol + str(default[int(round(Z))])]

    projectors, radii = par[:2]
    if len(par) == 3:
        extra = par[2]
    else:
        extra = {}

    if args.configuration:
        configuration = eval(args.configuration)
    else:
        configuration = None

    if args.projectors:
        projectors = args.projectors

    if args.radius:
        radii = [float(r) for r in args.radius.split(',')]

    if isinstance(radii, float):
        radii = [radii]

    if args.pseudize:
        type, nderiv = args.pseudize.split(',')
        pseudize = (type, int(nderiv))
    else:
        pseudize = ('poly', 4)

    if args.zero_potential:
        x = args.zero_potential.split(',')
        nderiv0 = int(x[0])
        r0 = float(x[1])
    else:
        if projectors[-1].isupper():
            nderiv0 = 5
            r0 = extra.get('r0', min(radii) * 0.9)
        else:
            nderiv0 = 2
            r0 = extra.get('r0', min(radii))

    if args.pseudo_core_density_radius:
        rcore = args.pseudo_core_density_radius
    else:
        rcore = extra.get('rcore')

    if args.nlcc:
        rcore *= -1

    return dict(symbol=symbol,
                Z=Z,
                xc=args.xc_functional,
                configuration=configuration,
                projectors=projectors,
                radii=radii,
                scalar_relativistic=args.scalar_relativistic, alpha=args.alpha,
                r0=r0, v0=None, nderiv0=nderiv0,
                pseudize=pseudize, rcore=rcore,
                core_hole=args.core_hole,
                yukawa_gamma=args.gamma,
                omega=args.omega)


def generate(symbol,
             Z,
             projectors,
             radii,
             r0, v0,
             nderiv0,
             xc='LDA',
             scalar_relativistic=False,
             pseudize=('poly', 4),
             configuration=None,
             alpha=None,
             rcore=None,
             core_hole=None,
             *,
             yukawa_gamma=0.0,
             omega=None):
    aea = AllElectronAtom(symbol, xc, Z=Z,
                          configuration=configuration)
    gen = PAWSetupGenerator(aea, projectors, scalar_relativistic, core_hole,
                            yukawa_gamma=yukawa_gamma, omega=omega)

    gen.construct_shape_function(alpha, radii, eps=1e-10)
    gen.calculate_core_density()
    gen.find_local_potential(r0, nderiv0, v0)
    gen.add_waves(radii)
    gen.pseudize(pseudize[0], pseudize[1], rcore=rcore)
    gen.construct_projectors(rcore)
    return gen


def generate_all():
    if len(sys.argv) > 1:
        atoms = sys.argv[1:]
    else:
        atoms = None
    functionals = ['LDA', 'PBE', 'revPBE', 'RPBE']
    for name in parameters:
        n = sum(c.isalpha() for c in name)
        symbol = name[:n]
        electrons = name[n:]
        if atoms and symbol not in atoms:
            continue
        print(name, symbol, electrons)
        for xc in functionals:
            argv = [symbol, '-swf', xc, '-e', electrons, '-t', electrons + 'e']
            if xc == 'PBE':
                argv.append('-b')
            main(argv)


class CLICommand:
    """Create PAW dataset."""

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('symbol')
        add('-f', '--xc-functional', type=str, default='LDA',
            help='Exchange-Correlation functional (default value LDA)',
            metavar='<XC>')
        add('-C', '--configuration',
            help='e.g. for Li: "[(1, 0, 2, -1.878564), (2, 0, 1, -0.10554),'
            ' (2, 1, 0, 0.0)]"')
        add('-P', '--projectors',
            help='Projector functions - use comma-separated - ' +
            'nl values, where n can be principal quantum number ' +
            '(integer) or energy (floating point number). ' +
            'Example: 2s,0.5s,2p,0.5p,0.0d.')
        add('-r', '--radius',
            help='1.2 or 1.2,1.1,1.1')
        add('-0', '--zero-potential',
            metavar='nderivs,radius',
            help='Parameters for zero potential.')
        add('-c', '--pseudo-core-density-radius', type=float,
            metavar='radius',
            help='Radius for pseudizing core density.')
        add('-z', '--pseudize',
            metavar='type,nderivs',
            help='Parameters for pseudizing wave functions.')
        add('-p', '--plot', action='store_true')
        add('-l', '--logarithmic-derivatives',
            metavar='spdfg,e1:e2:de,radius',
            help='Plot logarithmic derivatives. ' +
            'Example: -l spdf,-1:1:0.05,1.3. ' +
            'Energy range and/or radius can be left out.')
        add('-w', '--write', action='store_true')
        add('-s', '--scalar-relativistic', action='store_true')
        add('-n', '--no-check', action='store_true')
        add('-t', '--tag', type=str)
        add('-a', '--alpha', type=float)
        add('-g', '--gamma', type=float, default=0.0)
        add('-b', '--create-basis-set', action='store_true')
        add('--nlcc', action='store_true',
            help='Use NLCC-style pseudo core density '
            '(for vdW-DF functionals).')
        add('--core-hole')
        add('-e', '--electrons', type=int)
        add('--ri', type=str,
            help='Calculate also resolution of identity basis.')
        add('--omega', type=float, default=None,
            help='Calculate core-core and core-valence contributions'
                 ' to erfc screen HSE-potential')

    @staticmethod
    def run(args):
        main(args)


def main(args):
    kwargs = get_parameters(args.symbol, args)
    gen = generate(**kwargs)

    if not args.no_check:
        if not gen.check_all():
            raise DatasetGenerationError

    if args.create_basis_set or args.write:
        if args.create_basis_set:
            basis = gen.create_basis_set(tag=args.tag, ri=args.ri)
            basis.write_xml()

        if args.write:
            setup = gen.make_paw_setup(args.tag)
            parameters = []
            for key, value in kwargs.items():
                if value is not None:
                    parameters.append('{0}={1!r}'.format(key, value))
            setup.generatordata = ',\n    '.join(parameters)
            setup.write_xml()

    if not args.create_basis_set and args.ri:
        raise ValueError('Basis set must be created in order to create the '
                         'RI-basis set as well')

    if args.logarithmic_derivatives or args.plot:
        if args.plot:
            import matplotlib.pyplot as plt
        if args.logarithmic_derivatives:
            plot_log_derivs(gen, args.logarithmic_derivatives, args.plot)

        if args.plot:
            gen.plot()

            if args.create_basis_set:
                gen.basis.generatordata = ''  # we already printed this
                BasisPlotter(show=True).plot(gen.basis)

        if args.plot:
            try:
                plt.show()
            except KeyboardInterrupt:
                pass


def plot_log_derivs(gen, ld_str: str, plot: bool):
    """Make nice log-derivs plot."""
    import matplotlib.pyplot as plt
    r = 1.1 * gen.rcmax
    emin = min(min(wave.e_n) for wave in gen.waves_l) - 0.8
    emax = max(max(wave.e_n) for wave in gen.waves_l) + 0.8
    lvalues, energies, r = parse_ld_str(ld_str, (emin, emax, 0.05), r)
    emin = energies[0]
    de = energies[1] - emin

    error = 0.0
    for l in lvalues:
        efix = []
        # Fixed points:
        if l < len(gen.waves_l):
            efix.extend(gen.waves_l[l].e_n)
        if l == gen.l0:
            efix.append(0.0)

        ld1 = gen.aea.logarithmic_derivative(l, energies, r)
        ld2 = gen.logarithmic_derivative(l, energies, r)
        for e in efix:
            i = int((e - emin) / de)
            if 0 <= i < len(energies):
                ld1 -= round(ld1[i] - ld2[i])
                if plot:
                    ldfix = ld1[i]
                    plt.plot([energies[i]], [ldfix],
                             'x' + colors[l])

        if plot:
            plt.plot(energies, ld1, colors[l], label='spdfg'[l])
            plt.plot(energies, ld2, '--' + colors[l])

        error = abs(ld1 - ld2).sum() * de
        print('Logarithmic derivative error:', l, error)

    if plot:
        plt.xlabel('energy [Ha]')
        plt.ylabel(r'$\arctan(d\log\phi_{\ell\epsilon}(r)/dr)/\pi'
                   r'|_{r=r_c}$')
        plt.legend(loc='best')



import pytest
import numpy.random as ra
from gpaw.setup import create_setup
from gpaw.xc import XC


@pytest.mark.ci
def test_radial_integral4():
    ra.seed(8)
    xc = XC('LDA')
    s = create_setup('H', xc)
    ni = s.ni
    nii = ni * (ni + 1) // 2
    D_p = 0.1 * ra.random((1, nii)) + 0.2

    def f(x):
        return x

    s.xc_correction.four_phi_integrals(D_p, f)

    # Check integrals using two_phi_integrals function and finite differences:
    pass



import sys
from pathlib import Path

from ase import Atoms
from ase.parallel import parprint

from .info import info
from gpaw import GPAW, PW, setup_paths
from gpaw.mpi import size


class CLICommand:
    """Test GPAW installation."""

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args):
        info()
        test()


def test():
    for path in setup_paths:
        if Path(path).is_dir():
            break
    else:
        print("""Could not find any atomic PAW-data or pseudopotentials!

You need to set the GPAW_SETUP_PATH environment variable to point to
the directories where PAW dataset and basis files are stored.  See
https://wiki.fysik.dtu.dk/gpaw/install.html#install-paw-datasets
for details.""", file=sys.stderr)
        return

    parprint(f'Doing a test calculation (cores: {size}): ... ',
             end='', flush=True)
    a = 2.5
    d = 0.9
    chain = Atoms('H', cell=[a, a, d], pbc=(False, False, True))
    chain.center()
    chain.calc = GPAW(mode=PW(200),
                      kpts=(1, 1, 8),
                      txt='test.txt')
    chain.get_forces()
    chain.get_stress()
    parprint('Done')
    if size == 1:
        print()
        print('Test parallel calculation with "gpaw -P 4 test".')



from ase.spacegroup import crystal
from gpaw import GPAW
from gpaw import PW
from gpaw.test import equal


def test_symmetry_fractional_translations_med():
    'quartz'
    # no. 152 - trigonal

    a = 5.032090
    c = a * 1.0968533
    p0 = (0.4778763, 0.0, 1. / 3.)
    p1 = (0.4153076, 0.2531340, 0.2029893)

    atoms = crystal(['Si', 'O'], basis=[p0, p1],
                    spacegroup=152, cellpar=[a, a, c, 90, 90, 120])

    # with fractional translations
    calc = GPAW(mode=PW(),
                xc='LDA',
                kpts=(3, 3, 3),
                nbands=42,
                symmetry={'symmorphic': False},
                gpts=(20, 20, 24),
                eigensolver='rmm-diis')

    atoms.calc = calc
    energy_fractrans = atoms.get_potential_energy()

    assert len(calc.wfs.kd.ibzk_kc) == 7
    assert len(calc.wfs.kd.symmetry.op_scc) == 6

    # without fractional translations
    calc = GPAW(mode=PW(),
                xc='LDA',
                kpts=(3, 3, 3),
                nbands=42,
                gpts=(20, 20, 24),
                eigensolver='rmm-diis')

    atoms.calc = calc
    energy_no_fractrans = atoms.get_potential_energy()

    assert len(calc.wfs.kd.ibzk_kc) == 10
    assert len(calc.wfs.kd.symmetry.op_scc) == 2

    equal(energy_fractrans, energy_no_fractrans, 1e-3)



"""PES using approximate LrTDDFT scheme.

"""
import numpy as np

from ase.units import Hartree

from gpaw.pes import BasePES
from gpaw.pes.state import State
from gpaw.utilities import packed_index

from numpy import sqrt, pi


class TDDFTPES(BasePES):

    def __init__(self, mother, excited_daughter, daughter=None,
                 shift=True, tolerance={}):
        self.tolerance = {
            'occupation': 1e-10,
            'magnetic': 2e-6,
            'grid': 0,
        }
        for key in tolerance.keys():
            if key not in self.tolerance:
                raise RuntimeError("Tolerance key '%s' not known."
                                   % key)
            self.tolerance[key] = tolerance[key]

        if excited_daughter.calculator is not None:
            self.c_d = excited_daughter.calculator
        else:
            self.c_d = daughter

        self.c_m = mother
        self.gd = self.c_m.wfs.gd
        self.lr_d = excited_daughter

        self.c_m.converge_wave_functions()
        self.c_d.converge_wave_functions()
        self.lr_d.diagonalize()

        self.check_systems()
        self.lr_d.jend = self.lr_d.kss[-1].j

        # Make good way for initialising these

        kmax = 0
        lmax = 0
        for kss in self.lr_d.kss:
            kmax = max(kmax, kss.i)
            lmax = max(lmax, kss.j)
        self.kmax = kmax + 1
        self.lmax = lmax + 1

        self.f = None
        self.be = None
        self.shift = shift

        def gs_orbitals(calc):
            indicees = []
            nbands = calc.get_number_of_bands()
            spin = (calc.get_number_of_spins() == 2)
            f_tolerance = (2 - spin) * self.tolerance['occupation']
            for kpt in calc.wfs.kpt_u:
                for i in range(nbands):
                    if kpt.f_n[i] > f_tolerance:
                        indicees.append(i + kpt.s * nbands)
                        if not spin:
                            indicees.append(i + nbands)
            return indicees
        self.gs_m = gs_orbitals(self.c_m)
        self.imax = len(self.gs_m)
        self.gs_d = gs_orbitals(self.c_d)

        if (len(self.gs_m) != len(self.gs_d) + 1):
            raise RuntimeError(('Mother valence %d does not correspond ' +
                                'to daughter valence %d. ' +
                                'Modify tolerance["occupation"] ?') %
                               (len(self.gs_m), len(self.gs_d)))

    def _calculate(self):

        self.ks_overlaps()
        self.single_overlaps()
        self.full_overlap_matrix()

        self._create_f()

    def ks_overlaps(self):
        """Evaluate KS overlaps of mother and daughter."""
        bands_m = self.c_m.get_number_of_bands()
        spin_m = self.c_m.get_number_of_spins() == 2
        bands_d = self.c_d.get_number_of_bands()
        spin_d = self.c_d.get_number_of_spins() == 2

        self.overlap = np.zeros((2 * bands_m, 2 * bands_d))
        for i_m in range(bands_m):
            for s_m in range(2):
                k_m = spin_m * s_m
                wf_m = self.c_m.wfs.kpt_u[k_m].psit_nG[i_m]

                for j_d in range(bands_d):
                    k_d = spin_d * s_m

                    wf_d = self.c_d.wfs.kpt_u[k_d].psit_nG[j_d]
                    me = self.gd.integrate(wf_m * wf_d)

                    i = s_m * bands_m + i_m
                    j = s_m * bands_d + j_d
                    self.overlap[i, j] = me + self._nuc_corr(i_m, j_d,
                                                             k_m, k_d)

    def single_overlaps(self):
        self.singles = np.zeros((self.imax, len(self.lr_d)))
        nbands_d = self.c_d.get_number_of_bands()

        for i, i_m in enumerate(self.gs_m):
            for kl, kss in enumerate(self.lr_d.kss):
                if kss.fij > self.tolerance['occupation']:
                    spin = kss.pspin

                    keep_row = list(self.gs_m)
                    keep_row.remove(i_m)

                    k_d = kss.i + spin * nbands_d
                    l_d = kss.j + spin * nbands_d
                    keep_col = list(self.gs_d)
                    keep_col.remove(k_d)
                    keep_col.append(l_d)

                    d_ikl = np.zeros((len(keep_row), len(keep_col)))

                    for col in range(len(keep_col)):
                        for row in range(len(keep_row)):
                            d_ikl[row, col] = self.overlap[keep_row[row],
                                                           keep_col[col]]

                    self.singles[i, kl] = np.linalg.det(d_ikl)

    def gs_gs_overlaps(self):
        """Evaluate overlap matrix of mother and daughter ground states.

        """
        g0 = np.zeros((self.imax))
        for i, i_m in enumerate(self.gs_m):
            keep_row = list(self.gs_m)
            keep_row.remove(i_m)

            keep_col = list(self.gs_d)
            d_i00 = np.zeros((len(keep_row), len(keep_col)))

            for col in range(len(keep_col)):
                for row in range(len(keep_row)):
                    d_i00[row, col] = self.overlap[keep_row[row],
                                                   keep_col[col]]

            g0[i] = (-1) ** (self.imax + i) * np.linalg.det(d_i00)
        return g0

    def full_overlap_matrix(self):
        """Full overlap matrix of mother and daughter many particle states.

        """
        self.g_Ii = np.zeros((len(self.lr_d) + 1, self.imax))
        self.g_Ii[0, :] = self.gs_gs_overlaps()

        for I in range(len(self.lr_d)):
            for i in range(self.imax):
                gi = 0
                for kl in range(len(self.lr_d)):
                    gi += self.lr_d[I].f[kl] * self.singles[i, kl]
                self.g_Ii[1 + I, i] = (-1.) ** (self.imax + i) * gi

    def _create_f(self):
        self.f = (self.g_Ii * self.g_Ii).sum(axis=1)

        if self.shift:
            shift = (self.c_d.get_potential_energy() -
                     self.c_m.get_potential_energy())
        else:
            shift = float(self.shift)

        self.be = (np.array([0] + list(self.lr_d.get_energies() * Hartree)) +
                   shift)

    def _nuc_corr(self, i_m, j_d, k_m, k_d):
        ma = 0.0

        for a, P_ni_m in self.c_m.wfs.kpt_u[k_m].P_ani.items():
            P_ni_d = self.c_d.wfs.kpt_u[k_d].P_ani[a]
            Pi_i = P_ni_m[i_m]
            Pj_i = P_ni_d[j_d]
            Delta_pL = self.c_m.wfs.setups[a].Delta_pL

            for i in range(len(Pi_i)):
                for j in range(len(Pj_i)):
                    pij = Pi_i[i] * Pj_i[j]
                    ij = packed_index(i, j, len(Pi_i))
                    ma += Delta_pL[ij, 0] * pij

        ma = self.gd.comm.sum(ma)
        return sqrt(4 * pi) * ma

    def check_systems(self):
        """Check that mother and daughter systems correspond to each other.

        """
        gtol = self.tolerance['grid']
        mtol = self.tolerance['magnetic']
        if (np.abs(self.c_m.wfs.gd.cell_cv -
                   self.c_d.wfs.gd.cell_cv) > gtol).any():
            raise RuntimeError('Not the same grid:' +
                               str(self.c_m.wfs.gd.cell_cv) + ' !=' +
                               str(self.c_d.wfs.gd.cell_cv))
        if (np.abs(self.c_m.wfs.gd.h_cv -
                   self.c_d.wfs.gd.h_cv) > gtol).any():
            raise RuntimeError('Not the same grid')
        if (self.c_m.atoms.positions != self.c_m.atoms.positions).any():
            raise RuntimeError('Not the same atomic positions')
        if np.abs(np.abs(self.c_m.get_magnetic_moment() -
                         self.c_d.get_magnetic_moment()) - 1) > mtol:
            print(self.c_m.get_magnetic_moment(),
                  self.c_d.get_magnetic_moment())
            print(self.tolerance)
            raise RuntimeError(('Mother (%g) ' %
                                self.c_m.get_magnetic_moment()) +
                               ('and daughter spin (%g) ' %
                                self.c_d.get_magnetic_moment()) +
                               'are not compatible')

    def Dyson_orbital(self, I):
        """Return the Dyson orbital corresponding to excition I."""
        if not hasattr(self, 'g'):
            self._calculate()
        if not hasattr(self, 'morbitals'):
            nbands = self.c_m.get_number_of_bands()
            spin = self.c_m.get_number_of_spins() == 2
            morbitals_ig = []
            for i in self.gs_m:
                k = int(i >= nbands) * spin
                i -= nbands * int(i >= nbands)
                morbitals_ig.append(self.c_m.wfs.kpt_u[k].psit_nG[i])
            self.morbitals_ig = np.array(morbitals_ig)

        dyson = State(self.gd)
        gridnn = self.gd.zeros()
        for i, g in enumerate(self.g_Ii[I]):
            gridnn += g * self.morbitals_ig[i]
        dyson.set_grid(gridnn)
        dyson.set_energy(-self.be[I])
        return dyson



# Copyright (C) 2006  CAMP
# Please see the accompanying LICENSE file for further information.

from math import pi, log, sqrt, factorial as fac

import numpy as np

"""Fourier filtering

This module is an implementation of this Fourier filtering scheme:

*A general and efficient pseudopotential Fourier filtering scheme for
real space methods using mask functions*, Maxim Tafipolsky, Rochus
Schmid, J Chem Phys. 2006 May 7;124:174102.

Only difference is that we use a gaussian for the mask function.  The
filtering is used for the projector functions and for the zero
potential."""

# 3D-Fourier transform:
#
#                 /     _ _
# ~         ^    |  _  iq.r          ^
# f (q) Y  (q) = | dr e    f (r) Y  (r)
#  l     lm      |          l     lm
#               /
#
# Radial part:
#
#                  /
#   ~        __ l |  2
#   f (q) = 4||i  | r dr j (qr) f (r)
#    l            |       l      l
#                /
#

# XXX use fast bessel transform !!!


class Filter:
    """Mask-function Fourier filter"""

    def __init__(self, r_g, dr_g, gcut, h):
        """Construct filter.

        The radial grid is defined by r(g) and dr/dg(g) (`r_g` and
        `dr_g`), `gcut` is the cutoff grid point, and `h` is the target
        grid spacing used in the calculation."""

        self.gcut = gcut
        rcut = r_g[gcut]

        N = 200
        self.r_g = r_g = r_g[:gcut].copy()  # will be modified later!
        self.dr_g = dr_g[:gcut]

        # Matrices for Bessel transform:
        q1 = 5 * pi / h / N
        self.q_i = q_i = q1 * np.arange(N)
        self.c = sqrt(2 * q1 / pi)
        self.sinqr_ig = np.sin(q_i[:, None] * r_g) * self.c
        self.cosqr_ig = np.cos(q_i[:, None] * r_g) * self.c

        # Cutoff function:
        qmax = pi / h
        alpha = 1.1
        qcut = qmax / alpha
        icut = 1 + int(qcut / q1)
        beta = 5 * log(10) / (alpha - 1.0)**2
        self.cut_i = np.ones(N)
        self.cut_i[icut:] = np.exp(
            -np.clip(beta * (q_i[icut:] / qcut - 1.0)**2, 0, 400))
        # self.cut_i[icut:] = np.exp(
        #     -np.clip(0, 400, beta * (q_i[icut:] / qcut - 1.0)**2))

        # Mask function:
        gamma = 3 * log(10) / rcut**2
        self.m_g = np.exp(-gamma * r_g**2)

        # We will need to divide by these two!  Remove zeros:
        q_i[0] = 1.0
        r_g[0] = 1.0

    def filter(self, f_g, l=0):
        """Filter radial function.

        The function to be filtered is::

          f(r)     ^
          ---- Y  (r)
           r    lm

        Output is::

                l     ^
          g(r) r  Y  (r),
                   lm

        where the filtered radial part ``g(r)`` is returned."""

        r_g = self.r_g
        q_i = self.q_i

        fdrim_g = f_g[:self.gcut] * self.dr_g / self.m_g / r_g

        #         sin(x)
        # j (x) = ------,
        #  0        x
        #
        #         sin(x)   cos(x)
        # j (x) = ------ - ------,
        #  1         2       x
        #           x
        #
        #           3    1            3
        # j (x) = (--- - -) sin(x) - --- cos(x),
        #  2         3   x             2
        #           x                 x
        #
        #          15    6             15   1
        # j (x) = (-- - ---) sin(x) - (-- - -) cos(x).
        #  3        4     2             3   x
        #          x     x             x
        #

        if l == 0:
            fq_i = np.dot(self.sinqr_ig, fdrim_g * r_g) * self.cut_i
            fr_g = np.dot(fq_i, self.sinqr_ig)
        elif l == 1:
            fq_i = np.dot(self.sinqr_ig, fdrim_g) / q_i
            fq_i -= np.dot(self.cosqr_ig, r_g * fdrim_g)
            fq_i[0] = 0.0
            fq_i *= self.cut_i
            fr_g = np.dot(fq_i / q_i, self.sinqr_ig) / r_g
            fr_g -= np.dot(fq_i, self.cosqr_ig)
        elif l == 2:
            fq_i = 3 * np.dot(self.sinqr_ig, fdrim_g / r_g) / q_i**2
            fq_i -= np.dot(self.sinqr_ig, fdrim_g * r_g)
            fq_i -= 3 * np.dot(self.cosqr_ig, fdrim_g) / q_i
            fq_i[0] = 0.0
            fq_i *= self.cut_i
            fr_g = 3 * np.dot(fq_i / q_i**2, self.sinqr_ig) / r_g**2
            fr_g -= np.dot(fq_i, self.sinqr_ig)
            fr_g -= 3 * np.dot(fq_i / q_i, self.cosqr_ig) / r_g
        elif l == 3:  # XXX This should be tested
            fq_i = 15 * np.dot(self.sinqr_ig, fdrim_g / r_g**2) / q_i**3
            fq_i -= 6 * np.dot(self.sinqr_ig, fdrim_g) / q_i
            fq_i -= 15 * np.dot(self.cosqr_ig, fdrim_g / r_g) / q_i**2
            fq_i += np.dot(self.cosqr_ig, r_g * fdrim_g)
            fq_i[0] = 0.0
            fq_i *= self.cut_i
            fr_g = 15 * np.dot(fq_i / q_i**3, self.sinqr_ig) / r_g**3
            fr_g -= 6 * np.dot(fq_i / q_i, self.sinqr_ig) / r_g
            fr_g -= 15 * np.dot(fq_i / q_i**2, self.cosqr_ig) / r_g**2
            fr_g += np.dot(fq_i, self.cosqr_ig)

        else:
            raise NotImplementedError

        a_g = np.zeros(len(f_g))
        a_g[:self.gcut] = fr_g * self.m_g / r_g**(l + 1)

        #            n
        #           2 n!     n
        # j (x) = --------- x   for  x << 1.
        #  n      (2n + 1)!
        #
        # This formula is used for finding the value of
        #
        #       -l
        # f(r) r    for r -> 0
        #
        c = 2.0**l * fac(l) / fac(2 * l + 1) * self.c
        a_g[0] = np.dot(fq_i, q_i**(l + 1)) * c

        return a_g


if __name__ == '__main__':
    rc = 1.1
    gamma = 1.95
    rc2 = rc * gamma
    M = 300
    beta = 0.3
    gcut = 1 + int(M * rc / (beta + rc))
    g_g = np.arange(M)
    r_g = beta * g_g / (M - g_g)
    drdg_g = beta * M / (M - g_g)**2

    x_g = r_g / rc
    p_g = 1 - x_g**2 * (3 - 2 * x_g)
    p_g[gcut:] = 0.0
    # p_g = np.exp(-np.clip(5.0 * r_g**2, 0, 400))

    h = 0.4
    f = Filter(r_g, drdg_g, rc2, h)
    pf0_g = f.filter(p_g)
    pf1_g = f.filter(p_g * r_g**1, 1)
    pf2_g = f.filter(p_g * r_g**2, 2)

    if 0:
        for i in range(200):
            print(5 * pi / h * i / 200, pf0_g[i], pf1_g[i], pf2_g[i])
    if 1:
        for r, p, pf0, pf1, pf2 in zip(r_g, p_g, pf0_g, pf1_g, pf2_g):
            print(r, p, pf0, pf1, pf2)
            if r > rc2:
                break



from math import exp, pi, sqrt
import numpy as np

from gpaw.gauss import Gauss
from gpaw.test import equal
from gpaw.utilities.folder import Folder, Lorentz, Voigt  # noqa

# Gauss and Lorentz functions


def test_spectrum():
    width = 0.5
    x = 1.5

    equal(Gauss(width).get(x),
          exp(- x**2 / 2 / width**2) / sqrt(2 * pi) / width,
          1.e-15)
    equal(Gauss(width).fwhm, width * np.sqrt(8 * np.log(2)), 1.e-15)
    equal(Lorentz(width).get(x),
          width / (x**2 + width**2) / pi,
          1.e-15)
    equal(Lorentz(width).fwhm, width * 2, 1.e-15)

    # folder function

    for func in [Gauss, Lorentz, Voigt]:
        folder = Folder(width, func(width).__class__.__name__)

        x = [0, 2]
        y = [[2, 0, 1], [1, 1, 1]]

        xl, yl = folder.fold(x, y, dx=.7)

        # check first value
        yy = np.dot(np.array(y)[:, 0], func(width).get(xl[0] - np.array(x)))
        equal(yl[0, 0], yy, 1.e-15)
